import datetime
import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import Auction, AuctionOrder, AuctionOrderBy, AuctionSearchType, AuctionStatus, BattlEyeTypeFilter, \
    BidType, \
    CharacterBazaar, \
    InvalidContent, PvpTypeFilter, \
    Sex, SkillFilter, \
    Vocation, VocationAuctionFilter

FILE_BAZAAR_CURRENT_EMPTY = "bazaar/tibiacom_history_empty.txt"
FILE_BAZAAR_CURRENT = "bazaar/tibiacom_current.txt"
FILE_BAZAAR_CURRENT_ALL_FILTERS = "bazaar/tibiacom_current_all_filters.txt"
FILE_BAZAAR_HISTORY = "bazaar/tibiacom_history.txt"
FILE_AUCTION_FINISHED = "bazaar/tibiacom_auction_finished.txt"
FILE_AUCTION_NOT_FOUND = "bazaar/tibiacom_auction_not_found.txt"


class TestBazaar(TestCommons, unittest.TestCase):
    def test_character_bazaar_from_content_current_no_filters_selected(self):
        bazaar = CharacterBazaar.from_content(self.load_resource(FILE_BAZAAR_CURRENT))

        self.assertIsNotNone(bazaar)
        self.assertEqual(300, bazaar.page)
        self.assertEqual(482, bazaar.total_pages)
        self.assertEqual(12031, bazaar.results_count)
        self.assertEqual(25, len(bazaar.entries))
        self.assertIsNotNone(bazaar.url)

        auction = bazaar.entries[0]
        self.assertEqual(30237, auction.auction_id)
        self.assertEqual(800, auction.bid)
        self.assertEqual(BidType.MINIMUM, auction.bid_type)
        self.assertIsNotNone(auction.character_url)
        self.assertEqual(0, len(auction.displayed_items))

        self.assertIsNotNone(bazaar.filters)
        self.assertIsNone(bazaar.filters.world)
        self.assertIsNone(bazaar.filters.pvp_type)
        self.assertIsNone(bazaar.filters.battleye)
        self.assertIsNone(bazaar.filters.vocation)
        self.assertIsNone(bazaar.filters.min_level)
        self.assertIsNone(bazaar.filters.max_level)
        self.assertIsNone(bazaar.filters.skill)
        self.assertIsNone(bazaar.filters.min_skill_level)
        self.assertIsNone(bazaar.filters.max_skill_level)
        self.assertEqual(AuctionOrder.LOWEST_EARLIEST, bazaar.filters.order)

    def test_character_bazaar_from_content_current_all_filters_selected(self):
        bazaar = CharacterBazaar.from_content(self.load_resource(FILE_BAZAAR_CURRENT_ALL_FILTERS))

        self.assertIsNotNone(bazaar)
        self.assertEqual(1, bazaar.page)
        self.assertEqual(4, bazaar.total_pages)
        self.assertEqual(92, bazaar.results_count)
        self.assertEqual(25, len(bazaar.entries))
        self.assertIsNotNone(bazaar.url)

        auction = bazaar.entries[0]
        self.assertEqual(82526, auction.auction_id)
        self.assertEqual(57000, auction.bid)
        self.assertEqual(BidType.MINIMUM, auction.bid_type)
        self.assertIsNotNone(auction.character_url)

        self.assertIsNotNone(bazaar.filters)
        self.assertEqual('Antica', bazaar.filters.world)
        self.assertEqual(PvpTypeFilter.OPEN_PVP, bazaar.filters.pvp_type)
        self.assertEqual(BattlEyeTypeFilter.PROTECTED, bazaar.filters.battleye)
        self.assertEqual(VocationAuctionFilter.KNIGHT, bazaar.filters.vocation)
        self.assertEqual(1, bazaar.filters.min_level)
        self.assertEqual(1000, bazaar.filters.max_level)
        self.assertEqual(SkillFilter.MAGIC_LEVEL, bazaar.filters.skill)
        self.assertEqual(1, bazaar.filters.min_skill_level)
        self.assertEqual(50, bazaar.filters.max_skill_level)
        self.assertEqual(AuctionOrderBy.SHIELDING, bazaar.filters.order_by)
        self.assertEqual(AuctionOrder.HIGHEST_LATEST, bazaar.filters.order)
        self.assertEqual(AuctionSearchType.ITEM_WILDCARD, bazaar.filters.search_type)

    def test_character_bazaar_from_content_empty(self):
        bazaar = CharacterBazaar.from_content(self.load_resource(FILE_BAZAAR_CURRENT_EMPTY))
        self.assertIsNotNone(bazaar)
        self.assertFalse(bazaar.entries)

    def test_character_bazaar_from_content_history(self):
        bazaar = CharacterBazaar.from_content(self.load_resource(FILE_BAZAAR_HISTORY))

        self.assertIsNotNone(bazaar)
        self.assertEqual(1, bazaar.page)
        self.assertEqual(1449, bazaar.total_pages)
        self.assertEqual(36219, bazaar.results_count)
        self.assertEqual(25, len(bazaar.entries))
        self.assertIsNotNone(bazaar.url)

        auction = bazaar.entries[0]
        self.assertEqual(325058, auction.auction_id)
        self.assertEqual(900, auction.bid)
        self.assertEqual("Rcrazy Illuminati", auction.name)
        self.assertEqual(255, auction.level)
        self.assertEqual("Celebra", auction.world)
        self.assertEqual(Vocation.MASTER_SORCERER, auction.vocation)
        self.assertEqual(Sex.MALE, auction.sex)
        self.assertEqual(BidType.WINNING, auction.bid_type)
        self.assertIsNotNone(auction.character_url)
        self.assertEqual(1, len(auction.displayed_items))
        self.assertEqual(143, auction.outfit.outfit_id)

        first_item = auction.displayed_items[0]
        self.assertEqual(1, first_item.count)
        self.assertEqual(25700, first_item.item_id)
        self.assertEqual("dream blossom staff", first_item.name)
        self.assertIsNotNone(first_item.image_url)

        self.assertIsNone(bazaar.filters)

    def test_character_bazaar_from_content_unrelated(self):
        """Testing parsing an unrelated tibia.com section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            CharacterBazaar.from_content(content)

    def test_auction_details_from_content_finished(self):
        auction = Auction.from_content(self.load_resource(FILE_AUCTION_FINISHED))

        self.assertIsNotNone(auction)

        # Listing box
        self.assertEqual("Vireloz", auction.name)
        self.assertIn(auction.name, auction.character_url)
        self.assertIn(str(auction.auction_id), auction.url)
        self.assertEqual(1161, auction.level)
        self.assertEqual(Vocation.ROYAL_PALADIN, auction.vocation)
        self.assertEqual(Sex.MALE, auction.sex)
        self.assertEqual("Wintera", auction.world)
        self.assertIsNotNone(auction.outfit)
        self.assertEqual(1322, auction.outfit.outfit_id)
        self.assertEqual(4, len(auction.displayed_items))
        self.assertEqual("gnome armor", auction.displayed_items[0].name)
        self.assertEqual("falcon coif", auction.displayed_items[1].name)
        self.assertEqual("pair of soulstalkers", auction.displayed_items[2].name)
        self.assertEqual("lion spangenhelm", auction.displayed_items[3].name)

        self.assertEqual(330000, auction.bid)
        self.assertEqual(BidType.MINIMUM, auction.bid_type)
        self.assertEqual(AuctionStatus.FINISHED, auction.status)

        self.assertEqual(11715, auction.hit_points)
        self.assertEqual(17385, auction.mana)
        self.assertEqual(23530, auction.capacity)
        self.assertEqual(1270, auction.speed)
        self.assertEqual(0, auction.blessings_count)
        self.assertEqual(23, auction.mounts_count)
        self.assertEqual(35, auction.outfits_count)
        self.assertEqual(16, auction.titles_count)

        self.assertEqual(8, len(auction.skills))
        self.assertEqual(128, auction.skills_map["Distance Fighting"].level)
        self.assertEqual(11.43, auction.skills_map["Distance Fighting"].progress)

        self.assertIsInstance(auction.creation_date, datetime.datetime)
        self.assertEqual(26006721711, auction.experience)
        self.assertEqual(41893, auction.gold)
        self.assertEqual(553, auction.achievement_points)
        self.assertIsNone(auction.regular_world_transfer_available_date)
        self.assertEqual(110, auction.available_charm_points)
        self.assertEqual(5800, auction.spent_charm_points)

        self.assertEqual(2, auction.daily_reward_streak)
        self.assertEqual(1494, auction.hunting_task_points)
        self.assertEqual(0, auction.permanent_hunting_task_slots)
        self.assertEqual(1, auction.permanent_prey_slots)
        self.assertEqual(1, auction.hirelings)
        self.assertEqual(3, auction.hireling_jobs)
        self.assertEqual(0, auction.hireling_outfits)

        self.assertIsNotNone(auction.items)
        self.assertEqual(76, len(auction.items.entries))
        self.assertEqual(8, auction.items.total_pages)
        self.assertEqual(567, auction.items.results)
        self.assertEqual(141, auction.items.get_by_name("cigar").item_id)
        self.assertEqual("cigar", auction.items.get_by_id(141).name)
        self.assertEqual(7, len(auction.items.search('backpack')))

        self.assertIsNotNone(auction.store_items)
        self.assertEqual(16, len(auction.store_items.entries))
        self.assertEqual(1, auction.store_items.total_pages)
        self.assertEqual(16, auction.store_items.results)
        self.assertEqual(23721, auction.store_items.get_by_name("gold pouch").item_id)
        self.assertEqual("gold pouch", auction.store_items.get_by_id(23721).name)
        self.assertEqual(2, len(auction.store_items.search('rune')))

        self.assertIsNotNone(auction.mounts)
        self.assertEqual(22, len(auction.mounts.entries))
        self.assertEqual(1, auction.mounts.total_pages)
        self.assertEqual(22, auction.mounts.results)
        self.assertEqual(387, auction.mounts.get_by_name("donkey").mount_id)
        self.assertEqual("Donkey", auction.mounts.get_by_id(387).name)
        self.assertEqual(1, len(auction.mounts.search('drag')))

        self.assertIsNotNone(auction.store_mounts)
        self.assertEqual(1, len(auction.store_mounts.entries))
        self.assertEqual(1, auction.store_mounts.total_pages)
        self.assertEqual(1, auction.store_mounts.results)
        self.assertEqual(906, auction.store_mounts.get_by_name("Wolpertinger").mount_id)
        self.assertEqual("Wolpertinger", auction.store_mounts.get_by_id(906).name)
        self.assertEqual(1, len(auction.store_mounts.search('Wolpertinger')))

        self.assertIsNotNone(auction.outfits)
        self.assertEqual(30, len(auction.outfits.entries))
        self.assertEqual(2, auction.outfits.total_pages)
        self.assertEqual(33, auction.outfits.results)
        self.assertEqual(151, auction.outfits.get_by_name("pirate").outfit_id)
        self.assertEqual('Glooth Engineer', auction.outfits.get_by_id(610).name)
        self.assertEqual(2, len(auction.outfits.search('demon')))

        self.assertIsNotNone(auction.store_outfits)
        self.assertEqual(2, len(auction.store_outfits.entries))
        self.assertEqual(1, auction.store_outfits.total_pages)
        self.assertEqual(2, auction.store_outfits.results)
        self.assertEqual(962, auction.store_outfits.get_by_name("retro warrior").outfit_id)
        self.assertEqual('Retro Warrior', auction.store_outfits.get_by_id(962).name)
        self.assertEqual(2, len(auction.store_outfits.search('retro')))

        self.assertIsNotNone(auction.familiars)
        self.assertEqual(1, len(auction.familiars.entries))
        self.assertEqual(1, auction.familiars.total_pages)
        self.assertEqual(1, auction.familiars.results)
        self.assertEqual(992, auction.familiars.get_by_name("emberwing").familiar_id)
        self.assertEqual('Emberwing', auction.familiars.get_by_id(992).name)
        self.assertEqual(1, len(auction.familiars.search('ember')))

        self.assertEqual(9, len(auction.blessings))
        self.assertEqual(18, len(auction.imbuements))
        self.assertEqual(8, len(auction.charms))
        self.assertEqual(0, len(auction.completed_cyclopedia_map_areas))
        self.assertEqual(16, len(auction.titles))
        self.assertEqual(217, len(auction.achievements))
        self.assertEqual(509, len(auction.bestiary_progress))
        self.assertEqual(205, len(auction.completed_bestiary_entries))

    def test_auction_details_from_content_finished_skip_details(self):
        auction = Auction.from_content(self.load_resource(FILE_AUCTION_FINISHED), skip_details=True)

        self.assertIsNotNone(auction)

        # Listing box
        self.assertEqual("Vireloz", auction.name)
        self.assertIn(auction.name, auction.character_url)
        self.assertIn(str(auction.auction_id), auction.url)
        self.assertEqual(1161, auction.level)
        self.assertEqual(Vocation.ROYAL_PALADIN, auction.vocation)
        self.assertEqual(Sex.MALE, auction.sex)
        self.assertEqual("Wintera", auction.world)
        self.assertIsNotNone(auction.outfit)
        self.assertEqual(1322, auction.outfit.outfit_id)
        self.assertEqual(4, len(auction.displayed_items))
        self.assertEqual("gnome armor", auction.displayed_items[0].name)
        self.assertEqual("falcon coif", auction.displayed_items[1].name)
        self.assertEqual("pair of soulstalkers", auction.displayed_items[2].name)
        self.assertEqual("lion spangenhelm", auction.displayed_items[3].name)

        self.assertEqual(330000, auction.bid)
        self.assertEqual(BidType.MINIMUM, auction.bid_type)
        self.assertEqual(AuctionStatus.FINISHED, auction.status)

    def test_auction_details_from_content_not_found(self):
        auction = Auction.from_content(self.load_resource(FILE_AUCTION_NOT_FOUND))

        self.assertIsNone(auction)

    def test_auction_details_from_content_unrelated(self):
        """Testing parsing an unrelated tibia.com section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            Auction.from_content(content)


"""Asynchronous Tibia.com client."""
import asyncio
import datetime
import json
import logging
import time
import typing

import aiohttp
import aiohttp_socks

import tibiapy
from tibiapy import Auction, AuctionFilters, CharacterBazaar, Leaderboard, Spell, SpellsSection, abc
from tibiapy.character import Character
from tibiapy.creature import Creature, CreatureEntry, CreaturesSection
from tibiapy.enums import BattlEyeHighscoresFilter, Category, HouseType, NewsCategory, \
    NewsType, VocationFilter
from tibiapy.errors import Forbidden, NetworkError, SiteMaintenanceError
from tibiapy.event import EventSchedule
from tibiapy.forum import BoardEntry, CMPostArchive, ForumAnnouncement, ForumBoard, ForumPost, ForumThread
from tibiapy.guild import Guild, GuildWars, GuildsSection
from tibiapy.highscores import Highscores
from tibiapy.house import House, HousesSection
from tibiapy.kill_statistics import KillStatistics
from tibiapy.news import News, NewsArchive
from tibiapy.tournament import Tournament, TournamentLeaderboard
from tibiapy.world import World, WorldOverview

__all__ = (
    "TibiaResponse",
    "Client",
)

# Tibia.com's cache for the community section is 5 minutes.
# This limit is not sent anywhere, so there's no way to automate it.
CACHE_LIMIT = 300

T = typing.TypeVar('T')

log = logging.getLogger("tibiapy")


class TibiaResponse(typing.Generic[T], abc.Serializable):
    """Represents a response from Tibia.com.

    Attributes
    ----------
    timestamp: :class:`datetime.datetime`
        The date and time when the page was fetched, in UTC.
    cached: :class:`bool`
        Whether the response is cached or it is a fresh response.
    age: :class:`int`
        The age of the cache in seconds.
    fetching_time: :class:`float`
        The time in seconds it took for Tibia.com to respond.
    parsing_time: :class:`float`
        The time in seconds it took for the response to be parsed into data.
    data: :class:`T`
        The data contained in the response.
    """

    def __init__(self, raw_response, data: T, parsing_time=None):
        self.timestamp: datetime.datetime = raw_response.timestamp
        self.cached: bool = raw_response.cached
        self.age: int = raw_response.age
        self.fetching_time = raw_response.fetching_time
        self.parsing_time = parsing_time
        self.data: T = data

    __slots__ = (
        'timestamp',
        'cached',
        'age',
        'fetching_time',
        'parsing_time',
        'data',
    )

    _serializable_properties = ("time_left", )

    def __repr__(self):
        return f"<{self.__class__.__name__}[{type(self.data).__name__}] timestamp={self.timestamp!r} " \
               f"fetching_time={self.fetching_time!r} parsing_time={self.fetching_time!r} cahed={self.cached!r} " \
               f"age={self.age!r}>"

    @property
    def time_left(self):
        """:class:`datetime.timedelta`: The time left for the cache of this response to expire."""
        if not self.age:
            return datetime.timedelta()
        return (datetime.timedelta(seconds=CACHE_LIMIT - self.age)
                - (datetime.datetime.now(datetime.timezone.utc) - self.timestamp))

    @property
    def seconds_left(self):
        """:class:`int`: The time left in seconds for this response's cache to expire."""
        return self.time_left.seconds


class RawResponse:
    def __init__(self, response: aiohttp.ClientResponse, fetching_time: float):
        self.timestamp = datetime.datetime.now(datetime.timezone.utc)
        self.fetching_time = fetching_time
        self.cached = response.headers.get("CF-Cache-Status") == "HIT"
        age = response.headers.get("Age")
        if age is not None and age.isnumeric():
            self.age = int(age)
        else:
            self.age = 0
        self.content = None

    def __repr__(self):
        return f"<{self.__class__.__name__} timestamp={self.timestamp!r} fetching_time={self.fetching_time!r} " \
               f"cached={self.cached!r} age={self.age!r}>"


class Client:
    """An asynchronous client that fetches information from Tibia.com.

    The client uses a :class:`aiohttp.ClientSession` to request the information.
    A single session is shared across all operations.

    If desired, a custom ClientSession instance may be passed, instead of creating a new one.

    .. versionadded:: 2.0.0

    .. versionchanged:: 3.0.0
        All methods return a :class:`TibiaResponse` instance, containing additional information such as cache age.

    Attributes
    ----------
    loop : :class:`asyncio.AbstractEventLoop`
        The event loop to use. The default one will be used if not defined.
    session: :class:`aiohttp.ClientSession`
        The client session that will be used for the requests. One will be created by default.
    proxy_url: :class:`str`
        The URL of the SOCKS proxy to use for requests.
        Note that if a session is passed, the SOCKS proxy won't be used and must be applied when creating the session.
    """

    def __init__(self, loop=None, session=None, *, proxy_url=None):
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop
        self._session_ready = asyncio.Event()
        if session is not None:
            self.session: aiohttp.ClientSession = session
            self._session_ready.set()
        else:
            self.loop.create_task(self._initialize_session(proxy_url))

    # region Private Methods

    async def _initialize_session(self, proxy_url=None):
        """Initialize the aiohttp session object."""
        headers = {
            'User-Agent': "Tibia.py/%s (+https://github.com/Galarzaa90/tibia.py)" % tibiapy.__version__,
            'Accept-Encoding': "deflate, gzip",
        }
        connector = aiohttp_socks.SocksConnector.from_url(proxy_url) if proxy_url else None
        self.session: aiohttp.ClientSession = aiohttp.ClientSession(
            loop=self.loop,
            headers=headers,
            connector=connector,
        )
        self._session_ready.set()

    @classmethod
    def _handle_status(cls, status_code: int, fetching_time=0.0):
        """Handle error status codes, raising exceptions if necessary."""
        if status_code < 400:
            return
        if status_code == 403:
            raise Forbidden("403 Forbidden: Might be getting rate-limited", fetching_time=fetching_time)
        else:
            raise NetworkError("Request error, status code: %d" % status_code, fetching_time=fetching_time)

    async def _request(self, method, url, data=None, headers=None, *, test=False):
        """Base request, handling possible error statuses.

        Parameters
        ----------
        method: :class:`str`
            The HTTP method to use for the request.
        url: :class:`str`
            The URL that will be requested.
        data: :class:`dict`
            A mapping representing the form-data to send as part of the request.
        headers: :class:`dict`
            A mapping representing the headers to send as part of the request.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`RawResponse`
            The raw response obtained from the server.

        Raises
        ------
        Forbidden:
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        await self._session_ready.wait()
        if test:
            url = url.replace("www.tibia.com", "www.test.tibia.com")
        init_time = time.perf_counter()
        try:
            async with self.session.request(method, url, data=data, headers=headers) as resp:
                diff_time = time.perf_counter()-init_time
                if "maintenance.tibia.com" in str(resp.url):
                    log.info("%s | %s | %s %s | maintenance.tibia.com", url, resp.method, resp.status, resp.reason)
                    raise SiteMaintenanceError("Tibia.com is down for maintenance.")
                log.info("%s | %s | %s %s | %dms", url, resp.method, resp.status, resp.reason, int(diff_time * 1000))
                self._handle_status(resp.status, diff_time)
                response = RawResponse(resp, diff_time)
                response.content = await resp.text()
                return response
        except aiohttp.ClientError as e:
            raise NetworkError(f"aiohttp.ClientError: {e}", e, time.perf_counter() - init_time)
        except aiohttp_socks.SocksConnectionError as e:
            raise NetworkError(f"aiohttp_socks.SocksConnectionError: {e}", e, time.perf_counter() - init_time)
        except UnicodeDecodeError as e:
            raise NetworkError(f'UnicodeDecodeError: {e}', e, time.perf_counter() - init_time)

    async def _fetch_all_pages(self, auction_id, paginator, item_type, *, test=False):
        """Fetch all the pages of an auction paginator.

        Parameters
        ----------
        auction_id: :class:`int`
            The id of the auction.
        paginator:
            The paginator object
        item_type: :class:`int`
            The item type.
        test: :class:`bool`
            Whether to request the test website instead.
        """
        if paginator is None or paginator.entry_class is None:
            return
        current_page = 2
        while current_page <= paginator.total_pages:
            content = await self._fetch_ajax_page(auction_id, item_type, current_page, test=test)
            if content:
                entries = Auction._parse_page_items(content, paginator.entry_class)
                paginator.entries.extend(entries)
            current_page += 1
        paginator.fully_fetched = True

    async def _fetch_ajax_page(self, auction_id, type_id, page, *, test=False):
        """Fetch an ajax page from the paginated summaries in the auction section.

        Parameters
        ----------
        auction_id: :class:`int`
            The id of the auction.
        type_id: :class:`int`
            The ID of the type of the catalog to check.
        page: :class:`int`
            The page number to fetch.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`str`:
            The HTML content of the obtained page.
        """
        headers = {"x-requested-with": "XMLHttpRequest"}
        page_response = await self._request("GET", f"https://www.tibia.com/websiteservices/handle_charactertrades.php?"
                                                   f"auctionid={auction_id}&"
                                                   f"type={type_id}&"
                                                   f"currentpage={page}",
                                            headers=headers,
                                            test=test)
        try:
            data = json.loads(page_response.content.replace("\x0a", " "))
        except json.decoder.JSONDecodeError:
            return None
        try:
            return data['AjaxObjects'][0]['Data']
        except KeyError:
            return None

    # endregion

    # region Bazaar
    async def fetch_current_auctions(self, page=1, filters=None, *, test=False):
        """Fetch the current auctions in the bazaar.

        .. versionadded:: 3.3.0

        Parameters
        ----------
        page: :class:`int`
            The desired page to display.
        filters: :class:`AuctionFilters`
            The filtering criteria to use.

        Returns
        -------
        :class:`TibiaResponse` of :class:`CharacterBazaar`
            The current auctions.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        ValueError
            If the page number is not 1 or greater.
        """
        if page <= 0:
            raise ValueError('page must be 1 or greater.')
        response = await self._request("GET", CharacterBazaar.get_current_auctions_url(page, filters), test=test)
        start_time = time.perf_counter()
        current_auctions = CharacterBazaar.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, current_auctions, parsing_time)

    async def fetch_auction_history(self, page=1, filters=None, *, test=False):
        """Fetch the auction history of the bazaar.

        .. versionadded:: 3.3.0

        Parameters
        ----------
        page: :class:`int`
            The page to display.
        filters: :class:`AuctionFilters`
            The filtering criteria to use.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`CharacterBazaar`
            The character bazaar containing the auction history.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        ValueError
            If the page number is not 1 or greater.
        """
        if page <= 0:
            raise ValueError('page must be 1 or greater.')
        response = await self._request("GET", CharacterBazaar.get_auctions_history_url(page, filters), test=test)
        start_time = time.perf_counter()
        auction_history = CharacterBazaar.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, auction_history, parsing_time)

    async def fetch_auction(self, auction_id, *, fetch_items=False, fetch_mounts=False, fetch_outfits=False,
                            skip_details=False, test=False):
        """Fetch an auction by its ID.

        .. versionadded:: 3.3.0

        Parameters
        ----------
        auction_id: :class:`int`
            The ID of the auction.
        fetch_items: :class:`bool`
            Whether to fetch all of the character's items. By default only the first page is fetched.
        fetch_mounts: :class:`bool`
            Whether to fetch all of the character's mounts. By default only the first page is fetched.
        fetch_outfits: :class:`bool`
            Whether to fetch all of the character's outfits. By default only the first page is fetched.
        skip_details: :class:`bool`, optional
            Whether to skip parsing the entire auction and only parse the information shown in lists. False by default.

            This allows fetching basic information like name, level, vocation, world, bid and status, shaving off some
            parsing time.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Auction`
            The auction matching the ID if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        ValueError
            If the auction id is not 1 or greater.
        """
        if auction_id <= 0:
            raise ValueError('auction_id must be 1 or greater.')
        response = await self._request("GET", Auction.get_url(auction_id), test=test)
        start_time = time.perf_counter()
        auction = Auction.from_content(response.content, auction_id, skip_details)
        parsing_time = time.perf_counter() - start_time
        if auction and not skip_details:
            if fetch_items:
                await self._fetch_all_pages(auction_id, auction.items, 0, test=test)
                await self._fetch_all_pages(auction_id, auction.store_items, 1, test=test)
            if fetch_mounts:
                await self._fetch_all_pages(auction_id, auction.mounts, 2, test=test)
                await self._fetch_all_pages(auction_id, auction.store_mounts, 3, test=test)
            if fetch_outfits:
                await self._fetch_all_pages(auction_id, auction.outfits, 4, test=test)
                await self._fetch_all_pages(auction_id, auction.store_outfits, 5, test=test)
        return TibiaResponse(response, auction, parsing_time)

    # endregion

    async def fetch_cm_post_archive(self, start_date, end_date, page=1, *, test=False):
        """Fetch the CM post archive.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        start_date: :class: `datetime.date`
            The start date to display.
        end_date: :class: `datetime.date`
            The end date to display.
        page: :class:`int`
            The desired page to display.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`CMPostArchive`
            The CM Post Archive.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        ValueError
            If the start_date is more recent than the end date or page number is not 1 or greater.
        """
        if start_date > end_date:
            raise ValueError("start_date cannot be more recent than end_date")
        if page <= 0:
            raise ValueError("page cannot be lower than 1.")
        response = await self._request("GET", CMPostArchive.get_url(start_date, end_date, page), test=test)
        start_time = time.perf_counter()
        cm_post_archive = CMPostArchive.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, cm_post_archive, parsing_time)

    async def fetch_event_schedule(self, month=None, year=None, *, test=False):
        """Fetch the event calendar. By default, it gets the events for the current month.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        month: :class:`int`
            The month of the events to display.
        year: :class:`int`
            The year of the events to display.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`EventSchedule`
            The event calendar.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        ValueError
            If only one of year or month are defined.
        """
        if (year is None and month is not None) or (year is not None and month is None):
            raise ValueError("both year and month must be defined or neither must be defined.")
        response = await self._request("GET", EventSchedule.get_url(month, year), test=test)
        start_time = time.perf_counter()
        calendar = EventSchedule.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, calendar, parsing_time)

    # region Forums
    async def fetch_forum_community_boards(self, *, test=False):
        """Fetch the forum's community boards.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of list of :class:`BoardEntry`
            The forum boards in the community section.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", BoardEntry.get_community_boards_url(), test=test)
        start_time = time.perf_counter()
        boards = BoardEntry.list_from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boards, parsing_time)

    async def fetch_forum_support_boards(self, *, test=False):
        """Fetch the forum's community boards.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of list of :class:`BoardEntry`
            The forum boards in the community section.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", BoardEntry.get_support_boards_url(), test=test)
        start_time = time.perf_counter()
        boards = BoardEntry.list_from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boards, parsing_time)

    async def fetch_forum_world_boards(self, *, test=False):
        """Fetch the forum's world boards.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of list of :class:`BoardEntry`
            The forum boards in the world section.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", BoardEntry.get_world_boards_url(), test=test)
        start_time = time.perf_counter()
        boards = BoardEntry.list_from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boards, parsing_time)

    async def fetch_forum_trade_boards(self, *, test=False):
        """Fetch the forum's trade boards.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of list of :class:`BoardEntry`
            The forum boards in the trade section.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", BoardEntry.get_trade_boards_url(), test=test)
        start_time = time.perf_counter()
        boards = BoardEntry.list_from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boards, parsing_time)

    async def fetch_forum_board(self, board_id, page=1, age=30, *, test=False):
        """Fetch a forum board with a given id.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        board_id : :class:`int`
            The id of the board.
        page: :class:`int`
            The page number to show.
        age: :class:`int`
            The maximum age in days of the threads to display.

            To show threads of all ages, use -1.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`ForumBoard`
            A response containing the forum, if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", ForumBoard.get_url(board_id, page, age), test=test)
        start_time = time.perf_counter()
        board = ForumBoard.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, board, parsing_time)

    async def fetch_forum_thread(self, thread_id, page=1, *, test=False):
        """Fetch a forum thread with a given id.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        thread_id : :class:`int`
            The id of the thread.
        page: :class:`int`
            The desired page to display, by default 1.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`ForumThread`
            A response containing the forum, if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", ForumThread.get_url(thread_id, page), test=test)
        start_time = time.perf_counter()
        thread = ForumThread.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, thread, parsing_time)

    async def fetch_forum_post(self, post_id, *, test=False):
        """Fetch a forum post with a given id.

        The thread that contains the post will be returned, containing the desired post in
        :py:attr:`ForumThread.anchored_post`.

        The displayed page will be the page where the post is located.

        .. versionadded:: 3.1.0

        Parameters
        ----------
        post_id : :class:`int`
            The id of the post.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`ForumThread`
            A response containing the forum, if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", ForumPost.get_url(post_id), test=test)
        start_time = time.perf_counter()
        thread = ForumThread.from_content(response.content)
        if thread:
            thread.anchored_post = next((p for p in thread.posts if p.post_id == post_id), None)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, thread, parsing_time)

    async def fetch_forum_announcement(self, announcement_id, *, test=False):
        """Fetch a forum announcement.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        announcement_id: :class:`int`
            The id of the desired announcement.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`ForumAnnouncement`
            The forum announcement, if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", ForumAnnouncement.get_url(announcement_id), test=test)
        start_time = time.perf_counter()
        announcement = ForumAnnouncement.from_content(response.content, announcement_id)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, announcement, parsing_time)

    # endregion

    # region Creatures
    async def fetch_boosted_creature(self, *, test=False):
        """Fetch today's boosted creature.

        .. versionadded:: 2.1.0
        .. versionchanged:: 4.0.0
            The return type of the data returned was changed to :class:`Creature`, previous type was removed.

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`CreatureEntry`
            The boosted creature of the day.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", NewsArchive.get_url(), test=test)
        start_time = time.perf_counter()
        boosted_creature = CreaturesSection.boosted_creature_from_header(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boosted_creature, parsing_time)

    async def fetch_library_creatures(self, *, test=False):
        """Fetch the creatures from the library section.

        .. versionadded:: 4.0.0

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`CreaturesSection`
            The creature's section in Tibia.com

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", CreaturesSection.get_url(), test=test)
        start_time = time.perf_counter()
        boosted_creature = CreaturesSection.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boosted_creature, parsing_time)

    async def fetch_creature(self, identifier, *, test=False):
        """Fetch a creature's information from the Tibia.com library.

        .. versionadded:: 4.0.0

        Parameters
        ----------
        identifier: :class:`str`
            The internal name of the race.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Creature`
            The creature's section in Tibia.com

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", Creature.get_url(identifier), test=test)
        start_time = time.perf_counter()
        boosted_creature = Creature.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, boosted_creature, parsing_time)

    # endregion

    async def fetch_character(self, name, *, test=False):
        """Fetch a character by its name from Tibia.com.

        Parameters
        ----------
        name: :class:`str`
            The name of the character.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Character`
            A response containing the character, if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", Character.get_url(name.strip()), test=test)
        start_time = time.perf_counter()
        char = Character.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, char, parsing_time)

    # region Guilds
    async def fetch_guild(self, name, *, test=False):
        """Fetch a guild by its name from Tibia.com.

        Parameters
        ----------
        name: :class:`str`
            The name of the guild. The case must match exactly.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Guild`
            A response containing the found guild, if any.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", Guild.get_url(name), test=test)
        start_time = time.perf_counter()
        guild = Guild.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, guild, parsing_time)

    async def fetch_guild_wars(self, name, *, test=False):
        """Fetch a guild's wars by its name from Tibia.com.

        .. versionadded:: 3.0.0

        Parameters
        ----------
        name: :class:`str`
            The name of the guild. The case must match exactly.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`GuildWars`
            A response containing the found guild's wars.

            If the guild doesn't exist, the displayed data will show a guild with no wars instead of indicating the
            guild doesn't exist.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", GuildWars.get_url(name), test=test)
        start_time = time.perf_counter()
        guild_wars = GuildWars.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, guild_wars, parsing_time)

    # endregion

    async def fetch_house(self, house_id, world, *, test=False):
        """Fetch a house in a specific world by its id.

        Parameters
        ----------
        house_id: :class:`int`
            The house's internal id.
        world: :class:`str`
            The name of the world to look for.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`House`
            The house if found, :obj:`None` otherwise.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", House.get_url(house_id, world), test=test)
        start_time = time.perf_counter()
        house = House.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, house, parsing_time)

    async def fetch_highscores_page(self, world=None, category=Category.EXPERIENCE, vocation=VocationFilter.ALL, page=1,
                                    battleye_type=None, pvp_types=None, *, test=False):
        """Fetch a single highscores page from Tibia.com.

        Notes
        -----
        It is not possible to use BattlEye or PvPType filters when requesting a specific world.

        Parameters
        ----------
        world: :class:`str`
            The world to search the highscores in.
        category: :class:`Category`
            The highscores category to search, by default Experience.
        vocation: :class:`VocationFilter`
            The vocation filter to use. No filter used by default.
        page: :class:`int`
            The page to fetch, by default the first page is fetched.
        battleye_type: :class:`BattlEyeFilter`
            The type of BattlEye protection to display results from.
        pvp_types: :class:`list` of :class:`PvpTypeFilter`
            The list of PvP types to filter the results for.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Highscores`
            The highscores information or :obj:`None` if not found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        ValueError
            If an invalid filter combination is passed.
        """
        pvp_types = pvp_types or []
        if world is not None and ((battleye_type and battleye_type != BattlEyeHighscoresFilter.ANY_WORLD) or pvp_types):
            raise ValueError("BattleEye and PvP type filters can only be used when fetching all worlds.")
        response = await self._request("GET", Highscores.get_url(world, category, vocation, page, battleye_type,
                                                                 pvp_types), test=test)
        start_time = time.perf_counter()
        highscores = Highscores.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, highscores, parsing_time)

    async def fetch_kill_statistics(self, world, *, test=False):
        """Fetch the kill statistics of a world from Tibia.com.

        Parameters
        ----------
        world: :class:`str`
            The name of the world.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`KillStatistics`
            The kill statistics of the world if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", KillStatistics.get_url(world), test=test)
        start_time = time.perf_counter()
        kill_statistics = KillStatistics.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, kill_statistics, parsing_time)

    async def fetch_leaderboard(self, world, rotation=None, page=1, *, test=False):
        """Fetch the leaderboards for a specific world and rotation.

        .. versionadded:: 5.0.0

        Parameters
        ----------
        world: :class:`str`
            The name of the world.
        rotation: :class:`int`
            The ID of the rotation. By default it will get the current rotation.
        page: :class:`int`
            The page to get.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Leaderboard`
            The leaderboards of the world if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", Leaderboard.get_url(world, rotation, page), test=test)
        start_time = time.perf_counter()
        leaderboard = Leaderboard.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, leaderboard, parsing_time)

    # region Worlds
    async def fetch_world(self, name, *, test=False):
        """Fetch a world from Tibia.com.

        Parameters
        ----------
        name: :class:`str`
            The name of the world.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`World`
            A response containig the he world's information if found, :obj:`None` otherwise.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", World.get_url(name), test=test)
        start_time = time.perf_counter()
        world = World.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, world, parsing_time)

    async def fetch_world_houses(self, world, town, house_type=HouseType.HOUSE, status=None, order=None, *, test=False):
        """Fetch the house list of a world and type.

        .. versionchanged:: 5.0.0
            The data attribute of the response contains an instance of :class:`HousesSection` instead.

        Parameters
        ----------
        world: :class:`str`
            The name of the world.
        town: :class:`str`
            The name of the town.
        house_type: :class:`HouseType`
            The type of building. House by default.
        status: :class:`HouseStatus`, optional
            The house status to filter results. By default no filters will be applied.
        order: :class:`HouseOrder`, optional
            The ordering to use for the results. By default they are sorted by name.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`HousesSection`
            A response containing the lists of houses meeting the criteria if found.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", HousesSection.get_url(world=world, town=town, house_type=house_type,
                                                                    status=status, order=order), test=test)
        start_time = time.perf_counter()
        world_houses = HousesSection.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, world_houses, parsing_time)

    async def fetch_world_guilds(self, world: str, *, test=False):
        """Fetch the list of guilds in a world from Tibia.com.

        If a world that does not exist is passed, the world attribute of the result will be :obj:`None`.
        If the world attribute is set, but the list is empty, it just means the world has no guilds.

        .. versionchanged:: 5.0.0
            The data attribute of the response contains an instance of :class:`GuildsSection` instead.

        Parameters
        ----------
        world: :class:`str`
            The name of the world.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`GuildsSection`
            A response containing the guilds section for the specified world.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", GuildsSection.get_url(world), test=test)
        start_time = time.perf_counter()
        guilds = GuildsSection.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, guilds, parsing_time)

    async def fetch_world_list(self, *, test=False):
        """Fetch the world overview information from Tibia.com.

        Parameters
        ----------
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`WorldOverview`
            A response containing the world overview information.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", WorldOverview.get_url(), test=test)
        start_time = time.perf_counter()
        world_overview = WorldOverview.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, world_overview, parsing_time)

    # endregion

    # region News
    async def fetch_news_archive(self, start_date, end_date, categories=None, types=None, *, test=False):
        """Fetch news from the archive meeting the search criteria.

        .. versionchanged:: 5.0.0
            The data attribute of the response contains an instance of :class:`NewsArchive` instead.

        Parameters
        ----------
        start_date: :class:`datetime.date`
            The beginning date to search dates in.
        end_date: :class:`datetime.date`
            The end date to search dates in.
        categories: `list` of :class:`NewsCategory`
            The allowed categories to show. If left blank, all categories will be searched.
        types : `list` of :class:`NewsType`
            The allowed news types to show. if unused, all types will be searched.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`NewsArchive`
            The news meeting the search criteria.

        Raises
        ------
        ValueError:
            If ``begin_date`` is more recent than ``to_date``.
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        if start_date > end_date:
            raise ValueError("start_date can't be more recent than end_date")
        form_data = NewsArchive.get_form_data(start_date, end_date, categories, types)
        response = await self._request("POST", NewsArchive.get_url(), form_data, test=test)
        start_time = time.perf_counter()
        news = NewsArchive.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, news, parsing_time)

    async def fetch_recent_news(self, days=30, categories=None, types=None, *, test=False):
        """Fetch all the published news in the last specified days.

        This is a shortcut for :meth:`fetch_news_archive`, to handle dates more easily.

        .. versionchanged:: 5.0.0
            The data attribute of the response contains an instance of :class:`NewsArchive` instead.

        Parameters
        ----------
        days: :class:`int`
            The number of days to search, by default 30.
        categories: `list` of :class:`NewsCategory`
            The allowed categories to show. If left blank, all categories will be searched.
        types : `list` of :class:`NewsType`
            The allowed news types to show. if unused, all types will be searched.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`NewsArchive`
            The news posted in the last specified days.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        end = datetime.date.today()
        begin = end - datetime.timedelta(days=days)
        return await self.fetch_news_archive(begin, end, categories, types, test=test)

    async def fetch_news(self, news_id, *, test=False):
        """Fetch a news entry by its id from Tibia.com.

        Parameters
        ----------
        news_id: :class:`int`
            The id of the news entry.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`News`
            The news entry if found, :obj:`None` otherwise.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", News.get_url(news_id), test=test)
        start_time = time.perf_counter()
        news = News.from_content(response.content, news_id)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, news, parsing_time)
    # endregion

    # region Spells
    async def fetch_spells(self, *, vocation=None, group=None, spell_type=None, premium=None, sort=None, test=False):
        """Fetch the spells section.

        Parameters
        ----------
        vocation: :class:`VocationSpellFilter`, optional
            The vocation to filter in spells for.
        group: :class:`SpellGroup`, optional
            The spell's primary cooldown group.
        spell_type: :class:`SpellType`, optional
            The type of spells to show.
        premium: :class:`bool`, optional
            The type of premium requirement to filter. :obj:`None` means any premium requirement.
        sort: :class:`SpellSorting`, optional
            The field to sort spells by.

        Returns
        -------
        :class:`TibiaResponse` of :class:`SpellsSection`
            The spells section with the results.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", SpellsSection.get_url(vocation=vocation, group=group,
                                                                    spell_type=spell_type, premium=premium,
                                                                    sort=sort), test=test)
        start_time = time.perf_counter()
        spells = SpellsSection.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, spells, parsing_time)

    async def fetch_spell(self, identifier, *, test=False):
        """Fetch a spell by its identifier.

        Parameters
        ----------
        identifier: :class:`str`
            The spell's identifier. This is usually the name of the spell in lowercase and with no spaces.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Spell`
            The spell if found, :obj:`None` otherwise.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", Spell.get_url(identifier), test=test)
        start_time = time.perf_counter()
        spells = Spell.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, spells, parsing_time)
    # endregion

    # region Tournaments
    async def fetch_tournament(self, tournament_cycle=0, *, test=False):
        """Fetch a tournament from Tibia.com.

        .. versionadded:: 2.5.0

        Parameters
        ----------
        tournament_cycle: :class:`int`
            The cycle of the tournament. if unspecified, it will get the currently running tournament.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`Tournament`
            The tournament if found, :obj:`None` otherwise.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", Tournament.get_url(tournament_cycle), test=test)
        start_time = time.perf_counter()
        tournament = Tournament.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, tournament, parsing_time)

    async def fetch_tournament_leaderboard(self, tournament_cycle, world, page=1, *, test=False):
        """Fetch a tournament leaderboard from Tibia.com.

        .. versionadded:: 2.5.0

        Parameters
        ----------
        tournament_cycle: :class:`int`
            The cycle of the tournament. if unspecified, it will get the currently running tournament.
        world: :class:`str`
            The name of the world to get the leaderboards for.
        page: :class:`int`
            The desired leaderboards page, by default 1 is used.
        test: :class:`bool`
            Whether to request the test website instead.

        Returns
        -------
        :class:`TibiaResponse` of :class:`TournamentLeaderboard`
            The tournament's leaderboard if found, :obj:`None` otherwise.

        Raises
        ------
        Forbidden
            If a 403 Forbidden error was returned.
            This usually means that Tibia.com is rate-limiting the client because of too many requests.
        NetworkError
            If there's any connection errors during the request.
        """
        response = await self._request("GET", TournamentLeaderboard.get_url(world, tournament_cycle, page), test=test)
        start_time = time.perf_counter()
        tournament_leaderboard = TournamentLeaderboard.from_content(response.content)
        parsing_time = time.perf_counter() - start_time
        return TibiaResponse(response, tournament_leaderboard, parsing_time)
    # endregion



"""Tibia.com parsing and fetching library."""
__version__ = '5.0.1a1'
__author__ = 'Allan Galarza'
__license__ = 'Apache-2.0 License'

import logging

from tibiapy import abc, enums, utils
from tibiapy.house import *
from tibiapy.character import *
from tibiapy.creature import *
from tibiapy.event import *
from tibiapy.enums import *
from tibiapy.errors import *
from tibiapy.forum import *
from tibiapy.guild import *
from tibiapy.highscores import *
from tibiapy.kill_statistics import *
from tibiapy.leaderboard import *
from tibiapy.news import *
from tibiapy.tournament import *
from tibiapy.world import *
from tibiapy.bazaar import *
from tibiapy.spell import *
from tibiapy.client import *

from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())



import datetime
import sys
import unittest.mock

import aiohttp
import asynctest
from aioresponses import aioresponses

import tibiapy
from tests.tests_bazaar import FILE_BAZAAR_CURRENT, FILE_BAZAAR_HISTORY, FILE_AUCTION_FINISHED
from tests.tests_character import FILE_CHARACTER_RESOURCE, FILE_CHARACTER_NOT_FOUND
from tests.tests_events import FILE_EVENT_CALENDAR
from tests.tests_forums import FILE_BOARD_THREAD_LIST, FILE_CM_POST_ARCHIVE_PAGES, FILE_WORLD_BOARDS
from tests.tests_guild import FILE_GUILD_FULL, FILE_GUILD_LIST
from tests.tests_highscores import FILE_HIGHSCORES_FULL
from tests.tests_house import FILE_HOUSE_FULL, FILE_HOUSE_LIST
from tests.tests_kill_statistics import FILE_KILL_STATISTICS_FULL
from tests.tests_leaderboards import FILE_LEADERBOARD_CURRENT
from tests.tests_news import FILE_NEWS_LIST, FILE_NEWS_ARTICLE
from tests.tests_tibiapy import TestCommons
from tests.tests_world import FILE_WORLD_FULL, FILE_WORLD_LIST
from tibiapy import BoardEntry, CharacterBazaar, Client, Character, CMPostArchive, ForumBoard, Guild, GuildsSection, \
    Highscores, \
    HouseType, \
    HousesSection, \
    Leaderboard, NewsArchive, VocationFilter, \
    Category, \
    House, HouseEntry, \
    GuildEntry, \
    KillStatistics, NewsEntry, News, World, WorldOverview, Forbidden, NetworkError, CreatureEntry, Auction, \
    EventSchedule


class TestClient(asynctest.TestCase, TestCommons):
    def setUp(self):
        self.client = Client()

    async def tearDown(self):
        await self.client.session.close()

    async def test_client_init_pass_session(self):
        """Testing creating an instance passing a session"""
        headers = {"User-Agent": "Python Unit Test"}
        session = aiohttp.ClientSession(headers=headers)
        client = Client(session=session)

        self.assertEqual(client.session._default_headers, headers)

        await client.session.close()

    @aioresponses()
    async def test_client_handle_errors(self, mock):
        """Testing error handling"""
        mock.get(WorldOverview.get_url(), status=403)
        with self.assertRaises(Forbidden):
            await self.client.fetch_world_list()

        mock.get(WorldOverview.get_url(), status=404)
        with self.assertRaises(NetworkError):
            await self.client.fetch_world_list()

        mock.get(NewsEntry.get_list_url(), exception=aiohttp.ClientError())
        with self.assertRaises(NetworkError):
            await self.client.fetch_world_list()

        mock.post(NewsEntry.get_list_url(), exception=aiohttp.ClientOSError())
        with self.assertRaises(NetworkError):
            await self.client.fetch_recent_news(30)

    @aioresponses()
    async def test_client_fetch_character(self, mock):
        """Testing fetching a character"""
        name = "Tschas"
        content = self.load_resource(FILE_CHARACTER_RESOURCE)
        mock.get(Character.get_url(name), status=200, body=content)
        character = await self.client.fetch_character(name)

        self.assertIsInstance(character.data, Character)

    @aioresponses()
    async def test_client_fetch_character_not_found(self, mock):
        """Testing fetching a non existent character"""
        name = "Nezune"
        content = self.load_resource(FILE_CHARACTER_NOT_FOUND)
        mock.get(Character.get_url(name), status=200, body=content)
        character = await self.client.fetch_character(name)

        self.assertIsNone(character.data)

    @aioresponses()
    async def test_client_fetch_guild(self, mock):
        """Testing fetching a guild"""
        name = "Vitam et Mortem"
        content = self.load_resource(FILE_GUILD_FULL)
        mock.get(Guild.get_url(name), status=200, body=content)
        guild = await self.client.fetch_guild(name)

        self.assertIsInstance(guild.data, Guild)

    @aioresponses()
    async def test_client_fetch_world_guilds(self, mock):
        """Testing fetching a world's guild list"""
        world = "Zuna"
        content = self.load_resource(FILE_GUILD_LIST)
        mock.get(GuildsSection.get_url(world), status=200, body=content)
        response = await self.client.fetch_world_guilds(world)
        guilds = response.data.entries

        self.assertIsInstance(guilds[0], GuildEntry)

    @aioresponses()
    async def test_client_fetch_highscores_page(self, mock):
        """Testing fetching a highscores page"""
        world = "Estela"
        category = Category.MAGIC_LEVEL
        vocations = VocationFilter.KNIGHTS
        content = self.load_resource(FILE_HIGHSCORES_FULL)
        mock.get(Highscores.get_url(world, category, vocations), status=200, body=content)
        highscores = await self.client.fetch_highscores_page(world, category, vocations)

        self.assertIsInstance(highscores.data, Highscores)

    @aioresponses()
    async def test_client_fetch_house(self, mock):
        """Testing fetching a house"""
        world = "Antica"
        house_id = 5236
        content = self.load_resource(FILE_HOUSE_FULL)
        mock.get(House.get_url(house_id, world), status=200, body=content)
        house = await self.client.fetch_house(house_id, world)

        self.assertIsInstance(house.data, House)

    @aioresponses()
    async def test_client_fetch_world_houses(self, mock):
        """Testing fetching a world's houses"""
        world = "Antica"
        city = "Edron"
        content = self.load_resource(FILE_HOUSE_LIST)
        mock.get(HousesSection.get_url(world=world, town=city, house_type=HouseType.HOUSE), status=200, body=content)
        houses = await self.client.fetch_world_houses(world, city)

        self.assertIsInstance(houses.data, HousesSection)
        self.assertIsInstance(houses.data.entries[0], HouseEntry)

    @aioresponses()
    async def test_client_fetch_kill_statistics(self, mock):
        """Testing fetching kill statistics"""
        world = "Antica"
        content = self.load_resource(FILE_KILL_STATISTICS_FULL)
        mock.get(KillStatistics.get_url(world), status=200, body=content)
        kill_statistics = await self.client.fetch_kill_statistics(world)

        self.assertIsInstance(kill_statistics.data, KillStatistics)

    @aioresponses()
    async def test_client_fetch_recent_news(self, mock):
        """Testing fetching recent nows"""
        content = self.load_resource(FILE_NEWS_LIST)
        mock.post(NewsArchive.get_url(), status=200, body=content)
        recent_news = await self.client.fetch_recent_news(30)

        self.assertIsInstance(recent_news.data, NewsArchive)
        self.assertIsInstance(recent_news.data.entries[0], NewsEntry)

    async def test_client_fetch_news_archive_invalid_dates(self):
        """Testing fetching news archive with invalid dates"""
        today = datetime.date.today()
        yesterday = today - datetime.timedelta(days=1)
        with self.assertRaises(ValueError):
            await self.client.fetch_news_archive(today, yesterday)

    @aioresponses()
    async def test_client_fetch_news(self, mock):
        """Testing fetch news"""
        news_id = 6000
        content = self.load_resource(FILE_NEWS_ARTICLE)
        mock.get(News.get_url(news_id), status=200, body=content)
        news = await self.client.fetch_news(news_id)

        self.assertIsInstance(news.data, News)

    @aioresponses()
    async def test_client_fetch_world(self, mock):
        """Testing fetching a world"""
        name = "Antica"
        content = self.load_resource(FILE_WORLD_FULL)
        mock.get(World.get_url(name), status=200, body=content)
        world = await self.client.fetch_world(name)

        self.assertIsInstance(world.data, World)

    @aioresponses()
    async def test_client_fetch_world_list(self, mock):
        """Testing fetching the world list"""
        content = self.load_resource(FILE_WORLD_LIST)
        mock.get(WorldOverview.get_url(), status=200, body=content)
        worlds = await self.client.fetch_world_list()

        self.assertIsInstance(worlds.data, WorldOverview)

    @aioresponses()
    async def test_client_fetch_boosted_creature(self, mock):
        """Testing fetching the boosted creature"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        mock.get(News.get_list_url(), status=200, body=content)
        creature = await self.client.fetch_boosted_creature()

        self.assertIsInstance(creature.data, CreatureEntry)

    @aioresponses()
    async def test_client_fetch_cm_post_archive(self, mock):
        """Testing fetching the CM Post Archive"""
        content = self.load_resource(FILE_CM_POST_ARCHIVE_PAGES)
        start_date = datetime.date.today()-datetime.timedelta(days=40)
        end_date = datetime.date.today()
        mock.get(CMPostArchive.get_url(start_date, end_date), status=200, body=content)
        cm_post_archive = await self.client.fetch_cm_post_archive(start_date, end_date)

        self.assertIsInstance(cm_post_archive.data, CMPostArchive)

    async def test_client_fetch_cm_post_archive_invalid_dates(self):
        """Testing fetching the CM Post Archive with invalid dates"""
        end_date = datetime.date.today()-datetime.timedelta(days=40)
        start_date = datetime.date.today()
        with self.assertRaises(ValueError):
            await self.client.fetch_cm_post_archive(start_date, end_date)

    async def test_client_fetch_cm_post_archive_invalid_page(self):
        """Testing fetching the CM Post Archive with invalid page number"""
        start_date = datetime.date.today()-datetime.timedelta(days=40)
        end_date = datetime.date.today()
        with self.assertRaises(ValueError):
            await self.client.fetch_cm_post_archive(start_date, end_date, -1)

    @aioresponses()
    async def test_client_fetch_current_auctions(self, mock):
        """Testing fetching the current auctions"""
        content = self.load_resource(FILE_BAZAAR_CURRENT)
        mock.get(CharacterBazaar.get_current_auctions_url(), status=200, body=content)
        response = await self.client.fetch_current_auctions()
        self.assertIsInstance(response.data, CharacterBazaar)

    async def test_client_fetch_current_auctions_invalid_page(self):
        """Testing fetching the current auctions with an invalid page"""
        with self.assertRaises(ValueError):
            await self.client.fetch_current_auctions(-1)

    @aioresponses()
    async def test_client_fetch_auction_history(self, mock):
        """Testing fetching the auction history"""
        content = self.load_resource(FILE_BAZAAR_HISTORY)
        mock.get(CharacterBazaar.get_auctions_history_url(), status=200, body=content)
        response = await self.client.fetch_auction_history()
        self.assertIsInstance(response.data, CharacterBazaar)

    async def test_client_fetch_auction_history_invalid_page(self):
        """Testing fetching the auction history with an incorrect page"""
        with self.assertRaises(ValueError):
            await self.client.fetch_auction_history(-1)

    @aioresponses()
    async def test_client_fetch_auction(self, mock):
        """Testing fetching an auction"""
        content = self.load_resource(FILE_AUCTION_FINISHED)
        mock.get(Auction.get_url(134), status=200, body=content)
        response = await self.client.fetch_auction(134)
        self.assertIsInstance(response.data, Auction)

    async def test_client_fetch_auction_invalid_id(self):
        """Testing fetching an auction with an invalid id"""
        with self.assertRaises(ValueError):
            await self.client.fetch_auction(-1)

    @aioresponses()
    async def test_client_fetch_event_calendar(self, mock):
        """Testing fetching the auction history"""
        content = self.load_resource(FILE_EVENT_CALENDAR)
        mock.get(EventSchedule.get_url(), status=200, body=content)
        response = await self.client.fetch_event_schedule()
        self.assertIsInstance(response.data, EventSchedule)

    async def test_client_fetch_event_calendar_invalid_params(self):
        """Testing fetching the auction history"""
        with self.assertRaises(ValueError):
            await self.client.fetch_event_schedule(3)

    @aioresponses()
    async def test_client_fetch_forum_community_boards(self, mock):
        """Testing fetching the community boards"""
        content = self.load_resource(FILE_WORLD_BOARDS)
        mock.get(BoardEntry.get_community_boards_url(), status=200, body=content)
        response = await self.client.fetch_forum_community_boards()
        self.assertIsInstance(response.data[0], BoardEntry)

    @aioresponses()
    async def test_client_fetch_forum_trade_boards(self, mock):
        """Testing fetching the trade boards"""
        content = self.load_resource(FILE_WORLD_BOARDS)
        mock.get(BoardEntry.get_trade_boards_url(), status=200, body=content)
        response = await self.client.fetch_forum_trade_boards()
        self.assertIsInstance(response.data[0], BoardEntry)

    @aioresponses()
    async def test_client_fetch_forum_support_boards(self, mock):
        """Testing fetching the support boards"""
        content = self.load_resource(FILE_WORLD_BOARDS)
        mock.get(BoardEntry.get_support_boards_url(), status=200, body=content)
        response = await self.client.fetch_forum_support_boards()
        self.assertIsInstance(response.data[0], BoardEntry)

    @aioresponses()
    async def test_client_fetch_forum_world_boards(self, mock):
        """Testing fetching the world boards"""
        content = self.load_resource(FILE_WORLD_BOARDS)
        mock.get(BoardEntry.get_world_boards_url(), status=200, body=content)
        response = await self.client.fetch_forum_world_boards()
        self.assertIsInstance(response.data[0], BoardEntry)

    @aioresponses()
    async def test_client_fetch_forum_board(self, mock):
        """Testing fetching a forum board"""
        content = self.load_resource(FILE_BOARD_THREAD_LIST)
        mock.get(BoardEntry.get_url(1), status=200, body=content)
        response = await self.client.fetch_forum_board(1)
        self.assertIsInstance(response.data, ForumBoard)

    @aioresponses()
    async def test_client_fetch_leaderboards(self, mock):
        """Testing fetching the leaderboards"""
        content = self.load_resource(FILE_LEADERBOARD_CURRENT)
        mock.get(Leaderboard.get_url("Antica"), status=200, body=content)
        response = await self.client.fetch_leaderboard("Antica")
        self.assertIsInstance(response.data, Leaderboard)

    @unittest.mock.patch("tibiapy.bazaar.Auction._parse_page_items")
    @unittest.skipIf(sys.version_info < (3, 8, 0), "AsyncMock was implemented in 3.8")
    async def test_client__fetch_all_pages_success(self, parse_page_items):
        """Testing internal method to fetch all pages of an auction item collection."""
        paginator = tibiapy.ItemSummary(page=1, total_pages=5)
        self.client._fetch_ajax_page = unittest.mock.AsyncMock()

        await self.client._fetch_all_pages(1, paginator, 1)

        self.assertEqual(4, self.client._fetch_ajax_page.await_count)
        self.assertEqual(4, parse_page_items.call_count)

    async def test_client__fetch_all_pages_none_input(self):
        """Testing internal method to fetch all pages of an auction item collection."""
        self.assertIsNone(await self.client._fetch_all_pages(1, None, 1))


"""Models related to the creatures section in the library."""
import re
import urllib.parse
from typing import List, Optional

import bs4

from tibiapy import abc
from tibiapy.errors import InvalidContent

__all__ = (
    "CreatureEntry",
    "CreaturesSection",
    "Creature",
)

from tibiapy.utils import parse_tibiacom_content, get_tibia_url

BOOSTED_ALT = "Today's boosted creature: "

HP_PATTERN = re.compile(r"have (\d+) hitpoints")
EXP_PATTERN = re.compile(r"yield (\d+) experience")
IMMUNE_PATTERN = re.compile(r"immune to ([^.]+)")
WEAK_PATTERN = re.compile(r"weak against ([^.]+)")
STRONG_PATTERN = re.compile(r"strong against ([^.]+)")
LOOT_PATTERN = re.compile(r"They carry (.*) with them.")
MANA_COST = re.compile(r"takes (\d+) mana")


class CreaturesSection(abc.Serializable):
    """Represents the creature's section in the Tibia.com library.

    Attributes
    ----------
    boosted_creature: :class:`CreatureEntry`
        The current boosted creature.
    creatures: list of :class:`CreatureEntry`
        The list of creatures in the library.
    """

    __slots__ = (
        "boosted_creature",
        "creatures",
    )

    def __init__(self, boosted_creature, creatures):
        self.boosted_creature: CreatureEntry = boosted_creature
        self.creatures: List[CreatureEntry] = creatures or []

    @classmethod
    def get_url(cls):
        """Get the URL to the Tibia.com library section.

        Returns
        -------
        :class:`str`:
            The URL to the Tibia.com library section.
        """
        return get_tibia_url("library", "creature")

    @classmethod
    def boosted_creature_from_header(cls, content):
        """Get the boosted creature from any Tibia.com page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of a Tibia.com page.

        Returns
        -------
        :class:`CreatureEntry`
            The boosted creature of the day.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a Tibia.com's page.
        """
        try:
            parsed_content = bs4.BeautifulSoup(content.replace('ISO-8859-1', 'utf-8'), "lxml",
                                               parse_only=bs4.SoupStrainer("div", attrs={"id": "RightArtwork"}))
            img = parsed_content.find("img", attrs={"id": "Monster"})
            name = img["title"].replace(BOOSTED_ALT, "").strip()
            image_url = img["src"]
            identifier = image_url.split("/")[-1].replace(".gif", "")
            return CreatureEntry(name, identifier)
        except TypeError as e:
            raise InvalidContent("content is not from Tibia.com", e)

    @classmethod
    def from_content(cls, content):
        """Create an instance of the class from the html content of the creature library's page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        -------
        :class:`Character`
            The character contained in the page.

        Raises
        ------
        InvalidContent
            If content is not the HTML of a creature library's page.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            boosted_creature_table = parsed_content.find("div", {"class": "TableContainer"})
            boosted_creature_text = boosted_creature_table.find("div", {"class": "Text"})
            if not boosted_creature_text or "Boosted" not in boosted_creature_text.text:
                return None
            boosted_creature_link = boosted_creature_table.find("a")
            url = urllib.parse.urlparse(boosted_creature_link["href"])
            query = urllib.parse.parse_qs(url.query)
            boosted_creature = CreatureEntry(boosted_creature_link.text, query["race"][0])

            list_table = parsed_content.find("div", style=lambda v: v and 'display: table' in v)
            entries_container = list_table.find_all("div", style=lambda v: v and 'float: left' in v)
            entries = []
            for entry_container in entries_container:
                name = entry_container.text.strip()
                link = entry_container.find("a")
                url = urllib.parse.urlparse(link["href"])
                query = urllib.parse.parse_qs(url.query)
                entries.append(CreatureEntry(name, query["race"][0]))
            return cls(boosted_creature, entries)
        except (AttributeError, ValueError) as e:
            raise InvalidContent("content is not the creature's library", e)


class CreatureEntry(abc.Serializable):
    """Represents a creature in the Library section.

    Attributes
    ----------
    name: :class:`str`
        The name of the creature, usually in plural, except for the boosted creature.
    identifier: :class:`str`
        The internal name of the creature's race. Used for links and images.
    """

    __slots__ = (
        "name",
        "identifier",
    )

    _serializable_properties = (
        "image_url",
    )

    def __init__(self, name, identifier=None):
        self.name: str = name
        self.identifier: str = identifier

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} identifier={self.identifier!r}>"

    @property
    def url(self):
        """:class:`str`: The URL to this creature's details."""
        return self.get_url(self.identifier)

    @property
    def image_url(self):
        """:class:`str`: The URL to this creature's image."""
        return f"https://static.tibia.com/images/library/{self.identifier}.gif"

    @classmethod
    def get_url(cls, identifier):
        """Get the URL to the creature's detail page on Tibia.com.

        Parameters
        ----------
        identifier: :class:`str`
            The race's internal name.

        Returns
        -------
        :class:`str`
            The URL to the detail page.
        """
        return get_tibia_url("library", "creatures", race=identifier)


class Creature(CreatureEntry):
    """Represents a creature's details on the Tibia.com library.

    Attributes
    ----------
    name: :class:`str`
        The name of the creature, in plural form.
    identifier: :class:`str`
        The race's internal name. Used for links and images.
    description: :class:`str`
        A description of the creature.
    hitpoints: :class:`int`
        The number of hitpoints the creature has.
    experience: :class:`int`
        The number of experience points given for killing this creature.
    immune_to: list of :class:`str`
        The elements this creature is immune to.
    weak_against: list of :class:`str`
        The elements this creature is weak against.
    strong_against: list of :class:`str`
        The elements this creature is strong against.
    loot: :class:`str`
        Some of the items this creature drops.
    mana_cost: :class:`int`, optional
        The mana neccessary to summon or convince this creature.
    summonable: :class:`bool`
        Whether this creature can be summoned or not.
    convinceable: :class:`bool`
        Whether this creature can be convinced or not.
    """

    _valid_elements = ["ice", "fire", "earth", "poison", "death", "holy", "physical", "energy"]

    __slots__ = (
        "name",
        "identifier",
        "description",
        "hitpoints",
        "experience",
        "immune_to",
        "weak_against",
        "strong_against",
        "loot",
        "mana_cost",
        "summonable",
        "convinceable",
    )

    def __init__(self, name, identifier, **kwargs):
        super().__init__(name, identifier)
        self.immune_to: List[str] = kwargs.get("immune_to", [])
        self.weak_against: List[str] = kwargs.get("weak_against", [])
        self.strong_against: List[str] = kwargs.get("strong_against", [])
        self.loot: str = kwargs.get("loot")
        self.mana_cost: Optional[int] = kwargs.get("mana_cost")
        self.summonable: bool = kwargs.get("summonable", False)
        self.convinceable: bool = kwargs.get("convinceable", False)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name!r} identifier={self.identifier!r}>"

    @classmethod
    def from_content(cls, content):
        """Create an instance of the class from the html content of the creature library's page.

        Parameters
        ----------
        content: :class:`str`
            The HTML content of the page.

        Returns
        -------
        :class:`Creature`
            The character contained in the page.
        """
        try:
            parsed_content = parse_tibiacom_content(content)
            pagination_container, content_container = \
                parsed_content.find_all("div", style=lambda v: v and 'position: relative' in v)
            title_container, description_container = content_container.find_all("div")
            title = title_container.find("h2")
            name = title.text.strip()

            img = title_container.find("img")
            img_url = img["src"]
            race = img_url.split("/")[-1].replace(".gif", "")
            creature = cls(name, race)

            paragraph_tags = description_container.find_all("p")
            paragraphs = [p.text for p in paragraph_tags]
            creature.description = "\n".join(paragraphs[:-2])
            hp_text = paragraphs[-2]
            creature._parse_hp_text(hp_text)

            exp_text = paragraphs[-1]
            creature._parse_exp_text(exp_text)
            return creature
        except ValueError:
            return None

    def _parse_exp_text(self, exp_text):
        """Parse the experience text, containing dropped loot and adds it to the creature.

        Parameters
        ----------
        exp_text: :class:`str`
            The text containing experience.
        """
        m = EXP_PATTERN.search(exp_text)
        if m:
            self.experience = int(m.group(1))
        m = LOOT_PATTERN.search(exp_text)
        if m:
            self.loot = m.group(1)

    def _parse_hp_text(self, hp_text):
        """Parse the text containing the creatures hitpoints, containing weaknesses, immunities and more and adds it.

        Parameters
        ----------
        hp_text: :class:`str`
            The text containing hitpoints.
        """
        m = HP_PATTERN.search(hp_text)
        if m:
            self.hitpoints = int(m.group(1))
        m = IMMUNE_PATTERN.search(hp_text)
        if m:
            self._parse_elements(self.immune_to, m.group(1))
        if "cannot be paralysed" in hp_text:
            self.immune_to.append("paralyze")
        m = WEAK_PATTERN.search(hp_text)
        if m:
            self._parse_elements(self.weak_against, m.group(1))
        m = STRONG_PATTERN.search(hp_text)
        if m:
            self._parse_elements(self.strong_against, m.group(1))
        m = MANA_COST.search(hp_text)
        if m:
            self.mana_cost = int(m.group(1))
            if "summon or convince" in hp_text:
                self.convinceable = True
                self.summonable = True
            if "cannot be summoned" in hp_text:
                self.convinceable = True
            if "cannot be convinced" in hp_text:
                self.summonable = True
        if "sense invisible" in hp_text:
            self.immune_to.append("invisible")

    @classmethod
    def _parse_elements(cls, collection, text):
        """Parse the elements found in a string, adding them to the collection.

        Parameters
        ----------
        collection: :class:`list`
            The collection where found elements will be added to.
        text: :class:`str`
            The text containing the elements.
        """
        for element in cls._valid_elements:
            if element in text:
                collection.append(element)



import datetime
import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import NewsEntry, News, InvalidContent, NewsArchive
from tibiapy.enums import NewsCategory, NewsType

FILE_NEWS_LIST = "news/tibiacom_list.txt"
FILE_NEWS_LIST_EMPTY = "news/tibiacom_list_empty.txt"
FILE_NEWS_EMPTY = "news/tibiacom_empty.txt"
FILE_NEWS_ARTICLE = "news/tibiacom_news.txt"
FILE_NEWS_TICKER = "news/tibiacom_news_ticker.txt"


class TestNews(TestCommons, unittest.TestCase):
    # region Tibia.com Tests
    def test_news_archive_from_content(self):
        """Testing parsing news"""
        content = self.load_resource(FILE_NEWS_LIST)
        news_archive = NewsArchive.from_content(content)

        self.assertEqual(datetime.date(2019, 3, 25), news_archive.start_date)
        self.assertEqual(datetime.date(2019, 5, 25), news_archive.end_date)
        self.assertEqual(3, len(news_archive.types))
        self.assertEqual(5, len(news_archive.categories))

        news_list = news_archive.entries
        self.assertGreater(len(news_list), 0)
        latest_news = news_list[0]
        self.assertIsInstance(latest_news, NewsEntry)
        self.assertIsInstance(latest_news.id, int)
        self.assertIsInstance(latest_news.category, NewsCategory)
        self.assertIsInstance(latest_news.type, NewsType)
        self.assertIsInstance(latest_news.date, datetime.date)
        self.assertIsNotNone(latest_news.url)
        self.assertEqual(latest_news.url, NewsEntry.get_url(latest_news.id))

    def test_news_archive_from_content_empty(self):
        """Testing parsing a news article that doesn't exist"""
        content = self.load_resource(FILE_NEWS_LIST_EMPTY)
        news_archive = NewsArchive.from_content(content)

        self.assertEqual(datetime.date(2019, 5, 23), news_archive.start_date)
        self.assertEqual(datetime.date(2019, 5, 25), news_archive.end_date)
        self.assertEqual(1, len(news_archive.types))
        self.assertEqual(5, len(news_archive.categories))

        self.assertEqual(0, len(news_archive.entries))

    def test_news_archive_from_content_unrelated(self):
        """Testing parsing an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            NewsArchive.from_content(content)

    def test_news_from_content_empty(self):
        """Testing parsing an empty news article"""
        content = self.load_resource(FILE_NEWS_EMPTY)
        news = News.from_content(content)
        self.assertIsNone(news)

    def test_news_from_content_unrelated(self):
        """Testing parsing an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            News.from_content(content)

    def test_news_from_content_ticker(self):
        """Testing parsing a news ticker"""
        content = self.load_resource(FILE_NEWS_TICKER)
        news = News.from_content(content)

        self.assertIsInstance(news, News)
        self.assertEqual(news.category, NewsCategory.TECHNICAL_ISSUES)
        self.assertIsInstance(news.date, datetime.date)
        self.assertEqual(news.title, "News Ticker")
        self.assertIsNone(news.thread_id)

    def test_news_from_content_article(self):
        """Testing parsing an article"""
        content = self.load_resource(FILE_NEWS_ARTICLE)
        news = News.from_content(content)

        self.assertIsInstance(news, News)
        self.assertEqual(news.category, NewsCategory.DEVELOPMENT)
        self.assertIsInstance(news.date, datetime.date)
        self.assertEqual(news.title, "Sign Up for the VANGUARD Tournament")
        self.assertEqual(news.thread_id, 4725194)

    # endregion



import datetime
import unittest

import tibiapy
from tests.tests_tibiapy import TestCommons
from tibiapy import CMPostArchive, ForumAnnouncement, ForumBoard, ForumThread, InvalidContent, LastPost, BoardEntry, \
    ThreadEntry, ThreadStatus

FILE_WORLD_BOARDS = "forums/tibiacom_section.txt"
FILE_SECTION_EMPTY_BOARD = "forums/tibiacom_section_empty_board.txt"
FILE_SECTION_EMPTY = "forums/tibiacom_section_empty.txt"
FILE_BOARD_THREAD_LIST = "forums/tibiacom_board.txt"
FILE_BOARD_EMPTY_THREAD_LIST = "forums/tibiacom_board_empty.txt"
FILE_BOARD_INVALID_PAGE = "forums/tibiacom_board_invalid_page.txt"
FILE_BOARD_GOLDEN_FRAMES = "forums/tibiacom_board_golden_frame.txt"
FILE_ANNOUNCEMENT = "forums/tibiacom_announcement.txt"
FILE_ANNOUNCEMENT_NOT_FOUND = "forums/tibiacom_announcement_not_found.txt"
FILE_THREAD = "forums/tibiacom_thread.txt"
FILE_THREAD_NOT_FOUND = "forums/tibiacom_thread_not_found.txt"
FILE_THREAD_INVALID_PAGE = "forums/tibiacom_thread_invalid_page.txt"
FILE_CM_POST_ARCHIVE_INITIAL = "forums/tibiacom_cmpostarchive_initial.txt"
FILE_CM_POST_ARCHIVE_NO_PAGES = "forums/tibiacom_cmpostarchive_no_pages.txt"
FILE_CM_POST_ARCHIVE_NO_RESULTS = "forums/tibiacom_cmpostarchive_no_results.txt"
FILE_CM_POST_ARCHIVE_PAGES = "forums/tibiacom_cmpostarchive_pages.txt"


class TestForum(TestCommons, unittest.TestCase):
    def test_listed_board_list_from_content_world_boards(self):
        content = self.load_resource(FILE_WORLD_BOARDS)

        boards = BoardEntry.list_from_content(content)

        self.assertEqual(77, len(boards))
        for i, board in enumerate(boards):
            with self.subTest(i=i):
                self.assertIsInstance(board, BoardEntry)
                self.assertIsNotNone(board.name)
                self.assertGreater(board.board_id, 0)
                self.assertGreater(board.posts, 0)
                self.assertGreater(board.threads, 0)
                self.assertIsInstance(board.last_post, LastPost)

    def test_listed_board_list_from_content_empty_board(self):
        content = self.load_resource(FILE_SECTION_EMPTY_BOARD)

        boards = BoardEntry.list_from_content(content)

        self.assertEqual(77, len(boards))
        for i, board in enumerate(boards):
            with self.subTest(i=i):
                self.assertIsInstance(board, BoardEntry)
                self.assertIsNotNone(board.name)
                self.assertGreater(board.board_id, 0)
                self.assertGreaterEqual(board.posts, 0)
                self.assertGreaterEqual(board.threads, 0)
                self.assertIsNotNone(board.url)

    def test_listed_board_list_from_content_empty_section(self):
        content = self.load_resource(FILE_SECTION_EMPTY)

        boards = BoardEntry.list_from_content(content)

        self.assertIsInstance(boards, list)

    def test_listed_board_list_from_content_unrelated_section(self):
        content = self.load_resource(self.FILE_UNRELATED_SECTION)

        with self.assertRaises(InvalidContent):
            BoardEntry.list_from_content(content)

    def test_forum_board_from_content(self):
        content = self.load_resource(FILE_BOARD_THREAD_LIST)

        board = ForumBoard.from_content(content)

        self.assertIsNotNone(board)
        self.assertEqual("Antica", board.name)
        self.assertEqual("World Boards", board.section)
        self.assertEqual(1, board.page)
        self.assertEqual(5, board.total_pages)
        self.assertEqual(25, board.board_id)
        self.assertEqual(30, len(board.threads))
        self.assertIsNotNone(board.url)
        self.assertIsNotNone(board.next_page_url)
        self.assertEqual(board.next_page_url, BoardEntry.get_url(board.board_id, board.page + 1, board.age))
        for i, thread in enumerate(board.threads):
            with self.subTest(i=i):
                self.assertIsInstance(thread, ThreadEntry)
                self.assertIsNotNone(thread.title)
                self.assertGreater(thread.thread_id, 0)
                self.assertGreaterEqual(thread.views, 0)
                self.assertGreaterEqual(thread.replies, 0)
                self.assertIsInstance(thread.last_post, LastPost)
                self.assertIsNotNone(thread.last_post.author_url)

        with self.assertRaises(ValueError):
            board.get_page_url(-1)

    def test_board_section_urls(self):
        self.assertIsNotNone(BoardEntry.get_community_boards_url())
        self.assertIsNotNone(BoardEntry.get_support_boards_url())
        self.assertIsNotNone(BoardEntry.get_world_boards_url())
        self.assertIsNotNone(BoardEntry.get_trade_boards_url())

    def test_forum_board_from_content_empty_threads(self):
        content = self.load_resource(FILE_BOARD_EMPTY_THREAD_LIST)

        board = ForumBoard.from_content(content)

        self.assertIsNotNone(board)
        self.assertEqual("Ysolera - Trade", board.name)
        self.assertEqual("Trade Boards", board.section)
        self.assertEqual(1, board.page)
        self.assertEqual(1, board.total_pages)
        self.assertEqual(146059, board.board_id)
        self.assertEqual(0, len(board.threads))
        self.assertIsNone(board.next_page_url)
        self.assertIsNone(board.previous_page_url)

    def test_forum_board_from_content_unrelated_section(self):
        content = self.load_resource(self.FILE_UNRELATED_SECTION)

        with self.assertRaises(InvalidContent):
            ForumBoard.from_content(content)

    def test_forum_board_from_content_invalid_page(self):
        content = self.load_resource(FILE_BOARD_INVALID_PAGE)

        board = ForumBoard.from_content(content)

        self.assertEqual("Antica", board.name)
        self.assertEqual("World Boards", board.section)
        self.assertEqual([], board.threads)
        self.assertEqual(0, board.board_id)

    def test_forum_board_from_content_golden_frame(self):
        content = self.load_resource(FILE_BOARD_GOLDEN_FRAMES)

        board = ForumBoard.from_content(content)

        self.assertEqual("Proposals (English Only)", board.name)
        self.assertEqual("Community Boards", board.section)
        self.assertEqual(30, len(board.threads))
        self.assertEqual(5, len(board.announcements))
        self.assertEqual(1798, board.total_pages)
        for i, thread in enumerate(board.threads):
            with self.subTest(i=i):
                self.assertTrue(thread.golden_frame)
                self.assertTrue(thread.status & ThreadStatus.HOT)
                self.assertTrue(thread.status & ThreadStatus.CLOSED)
                self.assertIsNotNone(thread.url)

    def test_forum_announcement_from_content(self):
        content = self.load_resource(FILE_ANNOUNCEMENT)

        announcement = ForumAnnouncement.from_content(content, 33)

        self.assertIsNotNone(announcement)
        self.assertEqual("Legal Disclaimer", announcement.title)
        self.assertEqual(33, announcement.announcement_id)
        self.assertEqual("Proposals (English Only)", announcement.board)
        self.assertEqual(10, announcement.board_id)
        self.assertEqual("Community Boards", announcement.section)
        self.assertEqual(12, announcement.section_id)
        self.assertIsNotNone(announcement.author)
        self.assertEqual("CM Mirade", announcement.author.name)
        self.assertEqual(2, announcement.author.level)
        self.assertEqual(159, announcement.author.posts)
        self.assertEqual("Vunira", announcement.author.world)
        self.assertEqual("Community Manager", announcement.author.position)
        self.assertIsNotNone(announcement.url)

    def test_forum_announcement_from_content_not_found(self):
        content = self.load_resource(FILE_ANNOUNCEMENT_NOT_FOUND)

        announcement = ForumAnnouncement.from_content(content, 34)

        self.assertIsNone(announcement)

    def test_forum_announcement_from_content_unrelated_section(self):
        content = self.load_resource(self.FILE_UNRELATED_SECTION)

        with self.assertRaises(InvalidContent):
            ForumAnnouncement.from_content(content, 34)

    def test_forum_thread_from_content(self):
        content = self.load_resource(FILE_THREAD)

        thread = ForumThread.from_content(content)

        self.assertEqual("News: Team Finder, Visualisation of Loot Lists", thread.title)
        self.assertEqual(4797985, thread.thread_id)
        self.assertEqual('Auditorium (English Only)', thread.board)
        self.assertEqual('Community Boards', thread.section)
        self.assertEqual(4796826, thread.previous_topic_number)
        self.assertEqual(4797838, thread.next_topic_number)
        self.assertEqual(1, thread.page)
        self.assertEqual(9, thread.total_pages)
        self.assertEqual(20, len(thread.posts))
        self.assertIsNotNone(thread.url)
        self.assertIsNone(thread.previous_page_url)
        self.assertIsNotNone(thread.next_page_url)
        self.assertIsNotNone(thread.previous_thread_url)
        self.assertIsNotNone(thread.next_thread_url)

        post = thread.posts[0]
        self.assertEqual("Skerio", post.author.name)
        self.assertEqual("Relania", post.author.world)
        self.assertEqual("Community Manager", post.author.position)
        self.assertEqual(316, post.author.posts)
        self.assertEqual("Mirade", post.edited_by)
        self.assertTrue(post.golden_frame)
        self.assertEqual(38969385, post.post_id)
        self.assertIsNotNone(post.url)

        with self.assertRaises(ValueError):
            thread.get_page_url(-1)

    def test_forum_thread_from_content_invalid_page(self):
        content = self.load_resource(FILE_THREAD_INVALID_PAGE)

        thread = ForumThread.from_content(content)

        self.assertEqual("News: Team Finder, Visualisation of Loot Lists", thread.title)
        self.assertEqual(0, thread.thread_id)
        self.assertEqual('Auditorium (English Only)', thread.board)
        self.assertEqual('Community Boards', thread.section)
        self.assertEqual(4796826, thread.previous_topic_number)
        self.assertEqual(4797838, thread.next_topic_number)
        self.assertEqual(9, thread.page)
        self.assertEqual(9, thread.total_pages)
        self.assertEqual(0, len(thread.posts))

    def test_forum_thread_from_content_not_found(self):
        content = self.load_resource(FILE_THREAD_NOT_FOUND)

        thread = ForumThread.from_content(content)

        self.assertIsNone(thread)

    def test_forum_thread_from_content_unrelated_section(self):
        content = self.load_resource(self.FILE_UNRELATED_SECTION)

        with self.assertRaises(InvalidContent):
            ForumThread.from_content(content)

    def test_cm_post_archive_from_content_initial(self):
        content = self.load_resource(FILE_CM_POST_ARCHIVE_INITIAL)

        cm_post_archive = CMPostArchive.from_content(content)

        self.assertIsNotNone(cm_post_archive)
        self.assertEqual(0, cm_post_archive.results_count)
        self.assertEqual(1, cm_post_archive.page)
        self.assertEqual(1, cm_post_archive.total_pages)

    def test_cm_post_archive_from_content_no_pages(self):
        content = self.load_resource(FILE_CM_POST_ARCHIVE_NO_PAGES)

        cm_post_archive = CMPostArchive.from_content(content)

        self.assertIsNotNone(cm_post_archive)
        self.assertEqual(5, cm_post_archive.results_count)
        self.assertEqual(1, cm_post_archive.page)
        self.assertEqual(1, cm_post_archive.total_pages)
        self.assertEqual(cm_post_archive.results_count, len(cm_post_archive.posts))

        post = cm_post_archive.posts[0]
        self.assertIsInstance(post, tibiapy.abc.BasePost)
        self.assertEqual('Auditorium (English Only)', post.board)
        self.assertEqual(38974254, post.post_id)
        self.assertEqual('Ticker Messages June 2020', post.thread_title)

    def test_cm_post_archive_from_content_no_results(self):
        content = self.load_resource(FILE_CM_POST_ARCHIVE_NO_RESULTS)

        cm_post_archive = CMPostArchive.from_content(content)

        self.assertIsNotNone(cm_post_archive)
        self.assertEqual(0, cm_post_archive.results_count)
        self.assertEqual(1, cm_post_archive.page)
        self.assertEqual(1, cm_post_archive.total_pages)
        self.assertEqual(cm_post_archive.results_count, len(cm_post_archive.posts))

    def test_cm_post_archive_from_content_pages(self):
        content = self.load_resource(FILE_CM_POST_ARCHIVE_PAGES)

        cm_post_archive = CMPostArchive.from_content(content)

        self.assertIsNotNone(cm_post_archive)
        self.assertIsNotNone(cm_post_archive.url)
        self.assertIsNotNone(cm_post_archive.previous_page_url)
        self.assertIsNone(cm_post_archive.next_page_url)
        self.assertEqual(8370, cm_post_archive.results_count)
        self.assertEqual(168, cm_post_archive.page)
        self.assertEqual(168, cm_post_archive.total_pages)
        self.assertEqual(20, len(cm_post_archive.posts))

    def test_cm_post_archive_from_content_unrelated_section(self):
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            CMPostArchive.from_content(content)

    def test_cm_post_archive_get_page_url_negative_page(self):
        with self.assertRaises(ValueError):
            cmpost_archive = CMPostArchive()
            cmpost_archive.get_page_url(-1)

    def test_cm_post_archive_get_url_missing_parameter(self):
        with self.assertRaises(TypeError):
            CMPostArchive.get_url(None, datetime.date.today())
        with self.assertRaises(TypeError):
            CMPostArchive.get_url(datetime.date.today(), None)

    def test_cm_post_archive_get_url_invalid_dates(self):
        with self.assertRaises(ValueError):
            CMPostArchive.get_url(datetime.date.today(), datetime.date.today()-datetime.timedelta(days=20))

    def test_cm_post_archive_get_url_invalid_page(self):
        with self.assertRaises(ValueError):
            CMPostArchive.get_url(datetime.date.today()-datetime.timedelta(days=20), datetime.date.today(), -2)



import unittest

import tibiapy
from tests.tests_tibiapy import TestCommons
from tibiapy import Spell, SpellsSection

FILE_SPELLS_SECTION = "library/spell_list_default.txt"
FILE_SPELL_RUNE = "library/spell_rune.txt"
FILE_SPELL_SECONDARY_GROUP = "library/spell_secondary_group.txt"


class TestSpell(TestCommons, unittest.TestCase):
    # region Tibia.com Tests
    def test_spells_section_from_content(self):
        """Testing parsing a boosted creature"""
        content = self.load_resource(FILE_SPELLS_SECTION)
        spells_section = SpellsSection.from_content(content)

        self.assertIsNotNone(spells_section)
        self.assertEqual(141, len(spells_section.entries))

    def test_spells_section_from_content_unrelated_section(self):
        """Testing parsing a boosted creature"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(tibiapy.InvalidContent):
            spells_section = SpellsSection.from_content(content)

    def test_spell_from_content_rune(self):
        """Testing parsing a rune spell."""
        content = self.load_resource(FILE_SPELL_RUNE)
        spell = Spell.from_content(content)

        self.assertIsNotNone(spell)
        self.assertEqual("Chameleon Rune", spell.name)
        self.assertEqual("adevo ina", spell.words)
        self.assertIn("Druid", spell.vocations)
        self.assertEqual("Support", spell.group.value)
        self.assertEqual("Rune", spell.spell_type.value)
        self.assertEqual(2, spell.cooldown)
        self.assertEqual(2, spell.cooldown_group)
        self.assertEqual(2, spell.soul_points)
        self.assertEqual(1, spell.amount)
        self.assertEqual(27, spell.exp_level)
        self.assertEqual(600, spell.mana)
        self.assertEqual(1300, spell.price)
        self.assertIn("Thais", spell.cities)
        self.assertIn("Yalahar", spell.cities)
        self.assertIn("Edron", spell.cities)
        self.assertFalse(spell.premium)
        self.assertEqual("Chameleon Rune", spell.rune.name)
        self.assertIn("Knight", spell.rune.vocations)
        self.assertEqual("Support", spell.rune.group.value)
        self.assertEqual(27, spell.rune.exp_level)
        self.assertEqual(4, spell.rune.magic_level)

    def test_spell_from_content_secondary_group(self):
        """Testing parsing a spell with a secondary group."""
        content = self.load_resource(FILE_SPELL_SECONDARY_GROUP)
        spell = Spell.from_content(content)

        self.assertIsNotNone(spell)
        self.assertEqual("Protector", spell.name)
        self.assertEqual("utamo tempo", spell.words)
        self.assertIn("Knight", spell.vocations)
        self.assertEqual("Support", spell.group.value)
        self.assertEqual("Focus", spell.group_secondary)
        self.assertEqual("Instant", spell.spell_type.value)
        self.assertEqual(2, spell.cooldown)
        self.assertEqual(2, spell.cooldown_group)
        self.assertEqual(2, spell.cooldown_group_secondary)
        self.assertIsNone(spell.soul_points)
        self.assertIsNone(spell.amount)
        self.assertEqual(55, spell.exp_level)
        self.assertEqual(200, spell.mana)
        self.assertEqual(6000, spell.price)
        self.assertIn("Edron", spell.cities)
        self.assertTrue(spell.premium)

    def test_spells_from_content_unknown_spell(self):
        """Testing parsing an unknown spell

        When trying to fetch a spell that doesn't exist, the website will just show the spells section."""
        content = self.load_resource(FILE_SPELLS_SECTION)
        spell = Spell.from_content(content)

        self.assertIsNone(spell)

    def test_spells_from_content_unrelated_section(self):
        """Testing parsing a boosted creature"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(tibiapy.InvalidContent):
            spell = Spell.from_content(content)

    # endregion



import datetime
import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import InvalidContent, Leaderboard, TournamentEntry, PvpType, RuleSet, ScoreSet, Tournament, \
    TournamentLeaderboard, \
    TournamentPhase

FILE_LEADERBOARD_CURRENT = "leaderboards/tibiacom_current_rotation.txt"


class TestLeaderboards(TestCommons, unittest.TestCase):
    # region Leaderboard Tests
    def test_leaderboard_from_content(self):
        """Testing parsing a leaderboard's page"""
        content = self.load_resource(FILE_LEADERBOARD_CURRENT)
        leaderboard = Leaderboard.from_content(content)

        self.assertIsNotNone(leaderboard)
        self.assertEqual("Antica", leaderboard.world)
        self.assertEqual(2, leaderboard.rotation.rotation_id)
        self.assertEqual(45, leaderboard.results_count)
        self.assertEqual(2, len(leaderboard.available_rotations))
        self.assertEqual(81, len(leaderboard.available_worlds))
        self.assertIsNotNone(leaderboard.url)
        self.assertIsNone(leaderboard.previous_page_url)
        self.assertIsNone(leaderboard.next_page_url)
        self.assertIsNotNone(leaderboard.get_page_url(2))

        first_entry = leaderboard.entries[0]
        self.assertEqual("Daso II", first_entry.name)
        self.assertEqual(1, first_entry.rank)
        self.assertEqual(80, first_entry.drome_level)

    def test_leaderboard_from_content_unrelated_section(self):
        """Testing parsing a leaderboard's page"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            Leaderboard.from_content(content)

    def test_leaderboard_get_url_invalid_page(self):
        with self.assertRaises(ValueError):
            Leaderboard.get_url("Antica", page=0)

    # endregion



"""These are functions used thorough the module that may not be intended for public use."""
import datetime
import functools
import itertools
import re
import urllib.parse
import warnings
from collections import OrderedDict
from typing import Dict, Optional, Tuple, Type, TypeVar, Union

import bs4

TIBIA_CASH_PATTERN = re.compile(r'(\d*\.?\d*)\s?k*$')


def convert_line_breaks(element):
    """Convert the <br> tags in a HTML elements to actual line breaks.

    Parameters
    ----------
    element: :class:`bs4.Tag`
        A BeautifulSoup object.
    """
    for br in element.find_all("br"):
        br.replace_with("\n")


def get_tibia_url(section, subtopic=None, *args, anchor=None, test=False, **kwargs):
    """Build a URL to Tibia.com with the given parameters.

    Parameters
    ----------
    section: :class:`str`
        The desired section (e.g. community, abouttibia, manual, library)
    subtopic: :class:`str`, optional
        The desired subtopic (e.g. characters, guilds, houses, etc)
    anchor: :class:`str`
        A link anchor to add to the link.
    args:
        A list of key-value pairs to add as query parameters.
        This allows passing multiple parameters with the same name.
    kwargs:
        Additional parameters to pass to the url as query parameters (e.g name, world, houseid, etc)
    test: :class:`bool`
        Whether to use the test website or not.

    Returns
    -------
    :class:`str`
        The generated Tibia.com URL.

    Examples
    --------
    >>> get_tibia_url("community", "houses", page="view", houseid=55302, world="Gladera")
    https://www.tibia.com/community/?subtopic=houses&page=view&houseid=55302&world=Gladera

    You can also build a dictionary and pass it like:

    >>> params = {'world': "Gladera"}
    >>> get_tibia_url("community", "worlds", **params)
    https://www.tibia.com/community/?subtopic=worlds&world=Gladera
    """
    base_url = "www.tibia.com" if not test else "www.test.tibia.com"
    url = f"https://{base_url}/{section}/?"
    params = OrderedDict(subtopic=subtopic) if subtopic else OrderedDict()
    if kwargs:
        for key, value in kwargs.items():
            if isinstance(value, str):
                value = value.encode('iso-8859-1')
            if value is None:
                continue
            params[key] = value
    url += urllib.parse.urlencode(params)
    if args:
        url += "&"
        url += urllib.parse.urlencode(args)
    if anchor:
        url += f"#{anchor}"
    return url


def parse_form_data(form: bs4.Tag, include_options=True):
    """Parse the currently selected values in a form.

    This should correspond to all the data the form would send if submitted.

    Parameters
    ----------
    form: :class:`bs4.Tag`
        A form tag.
    include_options: :class:`bool`
        Whether to also include listings of all the possible options.
        These will be nested inside the __options__ parameter of the resulting dictionary.

    Returns
    -------
    :class:`dict`
        A dictionary containing all the data.
    """
    data = {}
    if "action" in form.attrs:
        data["__action__"] = form.attrs["action"]
    if "method" in form.attrs:
        data["__method__"] = form.attrs["method"]
    text_inputs = form.find_all("input", {"type": "text"})
    data.update({field.attrs.get("name"): field.attrs.get("value") for field in text_inputs})
    selects = form.find_all("select")
    if include_options:
        data["__options__"] = {}
    for select in selects:
        name = select.attrs.get("name")
        selected_option = select.find("option", {"selected": True})
        if include_options:
            options = select.find_all("option")
            data["__options__"][name] = {opt.text: opt.attrs.get("value") for opt in options}
        data[name] = selected_option.attrs.get("value") if selected_option else None
    checkboxes = form.find_all("input", {"type": "checkbox", "checked": True})
    data.update({field.attrs.get("name"): field.attrs.get("value") for field in checkboxes})
    # Parse Radios
    all_radios = form.find_all("input", {"type": "radio"})
    for name, radios in itertools.groupby(all_radios, key=lambda t: t.attrs["name"]):
        selected_radio = next((r for r in radios if r.attrs.get("checked") is not None), None)
        if include_options:
            data["__options__"][name] = {str(r.next_sibling).strip(): r.attrs["value"] for r in radios}
        data[name] = selected_radio.attrs["value"] if selected_radio else None
    return data


def parse_integer(number: str, default: Optional[int] = 0):
    """Parse a string representing an integer, ignoring commas or periods.

    Parameters
    ----------
    number: :class:`str`
        A string representing a number.
    default: :class:`int`
        The default value to use if the string is not numeric.
        By default, 0 is used.

    Returns
    -------
    :class:`int`
        The represented integer, or the default value if invalid.
    """
    if number is None:
        return default
    try:
        number = re.sub(r'[,.]', '', number.strip())
        return int(number)
    except ValueError:
        return default


def parse_tibia_datetime(datetime_str) -> Optional[datetime.datetime]:
    """Parse date and time from the format used in Tibia.com.

    Accepted format:

    - ``MMM DD YYYY, HH:mm:ss ZZZ``, e.g. ``Dec 10 2018, 21:53:37 CET``.
    - ``MMM DD YYYY, HH:mm ZZZ``, e.g. ``Dec 10 2018, 21:53 CET``.

    Parameters
    -------------
    datetime_str: :class:`str`
        The date and time as represented in Tibia.com

    Returns
    -----------
    :class:`datetime.datetime`, optional
        The represented datetime, in UTC (timezone aware).
    """
    try:
        datetime_str = datetime_str.replace(",", "").replace("&#160;", " ").strip()
        # Extracting timezone
        tz = datetime_str[-4:].strip()

        # Convert time string to time object
        # Removing timezone cause CEST and CET are not supported
        try:
            t = datetime.datetime.strptime(datetime_str[:-4].strip(), "%b %d %Y %H:%M:%S")
        except ValueError:
            t = datetime.datetime.strptime(datetime_str[:-4].strip(), "%b %d %Y %H:%M")

        # Getting the offset
        if tz == "CET":
            utc_offset = 1
        elif tz == "CEST":
            utc_offset = 2
        else:
            return None
        # Add/subtract hours to get the real time
        t = t - datetime.timedelta(hours=utc_offset)
        return t.replace(tzinfo=datetime.timezone.utc)
    except (ValueError, AttributeError):
        return None


def parse_tibia_date(date_str) -> Optional[datetime.date]:
    """Parse a date from the format used in Tibia.com.

    Accepted format:

    - ``MMM DD YYYY``, e.g. ``Jul 23 2015``

    Parameters
    -----------
    date_str: :class:`str`
        The date as represented in Tibia.com

    Returns
    -----------
    :class:`datetime.date`, optional
        The represented date, in UTC (timezone aware).
    """
    try:
        t = datetime.datetime.strptime(date_str.strip(), "%b %d %Y")
        return t.date()
    except (ValueError, AttributeError):
        return None


def parse_tibia_forum_datetime(datetime_str, utc_offset=1):
    """Parse a date in the format used in the Tibia.com forums.

    Accepted format:

    - ``DD.MM.YY HH:mm:ss``, e.g. ``23.07.2015 21:30:30``

    Parameters
    ----------
    datetime_str: :class:`str`
        The string containing the date and time.
    utc_offset: :class:`int`
        The UTC offset to apply to the parsed datetime.

        Since the timestamps contain no timezone information, it can be passed as an additional parameter.

        By default CET (+1) is considered.

    Returns
    -------
    :class:`datetime`
        The represented datetime, in UTC (timezone aware).
    """
    t = datetime.datetime.strptime(datetime_str.strip(), "%d.%m.%Y %H:%M:%S")
    # Add/subtract hours to get the real time
    t = t - datetime.timedelta(hours=utc_offset)
    return t.replace(tzinfo=datetime.timezone.utc)


def parse_tibia_full_date(date_str) -> Optional[datetime.date]:
    """Parse a date in the fuller format used in Tibia.com.

    Accepted format:

    - ``MMMM DD, YYYY``, e.g. ``July 23, 2015``

    Parameters
    -----------
    date_str: :class:`str`
        The date as represented in Tibia.com

    Returns
    -----------
    :class:`datetime.date`, optional
        The represented date, in UTC (timezone aware).
    """
    try:
        t = datetime.datetime.strptime(date_str.strip(), "%B %d, %Y")
        return t.date()
    except (ValueError, AttributeError):
        return None


def parse_number_words(text_num):
    """Parse the word representation of a number to a integer.

    Parameters
    ----------
    text_num: :class:`str`
        The text representation of a number.

    Returns
    -------
    :class:`int`
        The number represented by the string.
    """
    numwords = {}
    units = [
        "zero", "one", "two", "three", "four", "five", "six", "seven", "eight",
        "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen",
        "sixteen", "seventeen", "eighteen", "nineteen",
    ]

    tens = ["", "", "twenty", "thirty", "forty", "fifty", "sixty", "seventy", "eighty", "ninety"]

    scales = ["hundred", "thousand", "million", "billion", "trillion"]

    numwords["and"] = (1, 0)
    for idx, word in enumerate(units):
        numwords[word] = (1, idx)
    for idx, word in enumerate(tens):
        numwords[word] = (1, idx * 10)
    for idx, word in enumerate(scales):
        numwords[word] = (10 ** (idx * 3 or 2), 0)

    current = result = 0
    text_num = text_num.replace("-", " ")
    for word in text_num.split():
        if word not in numwords:
            return 0

        scale, increment = numwords[word]
        current = current * scale + increment
        if scale > 100:
            result += current
            current = 0

    return result + current


def try_datetime(obj) -> Optional[datetime.datetime]:
    """Attempt to convert an object into a datetime.

    If the date format is known, it's recommended to use the corresponding function
    This is meant to be used in constructors.

    Parameters
    ----------
    obj: :class:`str`, :class:`dict`, :class:`datetime.datetime`
        The object to convert.

    Returns
    -------
    :class:`datetime.datetime`, optional
        The represented datetime, in UTC (timezone aware), or :obj:`None` if conversion wasn't possible.
    """
    if obj is None:
        return None
    if isinstance(obj, datetime.datetime):
        return obj
    return parse_tibia_datetime(obj)


def try_date(obj) -> Optional[datetime.date]:
    """Attempt to convert an object into a date.

    If the date format is known, it's recommended to use the corresponding function
    This is meant to be used in constructors.

    Parameters
    ----------
    obj: :class:`str`, :class:`datetime.datetime`, :class:`datetime.date`
        The object to convert.

    Returns
    -------
    :class:`datetime.date`, optional
        The represented date, in UTC (timezone aware).
    """
    if obj is None:
        return None
    if isinstance(obj, datetime.datetime):
        return obj.date()
    if isinstance(obj, datetime.date):
        return obj
    res = parse_tibia_date(obj)
    if res is not None:
        return res
    return parse_tibia_full_date(obj)


def parse_tibiacom_content(content, *, html_class="BoxContent", tag="div", builder="lxml"):
    """Parse HTML content from Tibia.com into a BeautifulSoup object.

    Parameters
    ----------
    content: :class:`str`
        The raw HTML content from Tibia.com
    html_class: :class:`str`
        The HTML class of the parsed element. The default value is ``BoxContent``.
    tag: :class:`str`
        The HTML tag select. The default value is ``div``.
    builder: :class:`str`
        The builder to use. The default value is ``lxml``.

    Returns
    -------
    :class:`bs4.BeautifulSoup`, optional
        The parsed content.
    """
    strainer = bs4.SoupStrainer(tag, class_=html_class) if builder != "html5lib" else None
    return bs4.BeautifulSoup(content.replace('ISO-8859-1', 'utf-8', 1), builder, parse_only=strainer)


def parse_tibiacom_tables(parsed_content) -> Dict[str, bs4.Tag]:
    """Parse tables from Tibia.com into a mapping by the tables title.

    This is used for the table style used in Tibia.com, where a table is wrapped in a container with a title.

    Parameters
    ----------
    parsed_content: :class:`bs4.BeautifulSoup`
        The content to find the tables in.

    Returns
    -------
    :class:`dict`
        A dictionary mapping the container titles and the contained table.
    """
    table_containers = parsed_content.find_all("div", attrs={"class": "TableContainer"})
    tables = {}
    for table_container in table_containers:
        text_tag = table_container.find("div", attrs={"class": "Text"})
        table = table_container.find("table", attrs={"class": "TableContent"})
        if not table:
            continue
        tables[text_tag.text.strip()] = table
    return tables


T = TypeVar('T')
D = TypeVar('D')


def try_enum(cls: Type[T], val, default: D = None) -> Union[T, D]:
    """Attempt to convert a value into their enum value.

    Parameters
    ----------
    cls: :class:`Enum`
        The enum to convert to.
    val:
        The value to try to convert to Enum
    default: optional
        The value to return if no enum value is found.

    Returns
    -------
    obj:
        The enum value if found, otherwise None.
    """
    if isinstance(val, cls):
        return val
    try:
        return cls(val)
    except ValueError:
        try:
            if isinstance(val, str):
                val = val.upper()
            return cls._member_map_[val]
        except KeyError:
            return default


def parse_tibia_money(argument):
    """Parse a string that may contain 'k' as thousand suffix.

    Parameters
    ----------
    argument: :class:`str`
        A numeric string using k as a prefix for thousands.

    Returns
    -------
    :class:`int`:
        The value represented by the string.
    """
    try:
        return int(argument)
    except ValueError:
        argument = argument.replace(",", "").strip().lower()
        m = TIBIA_CASH_PATTERN.match(argument)
        if not m or not m.group(1):
            raise ValueError("not a numeric value")
        num = float(m.group(1))
        k_count = argument.count("k")
        num *= pow(1000, k_count)
        return int(num)


def split_list(items, separator=",", last_separator=" and "):
    """Split a string listing elements into an actual list.

    Parameters
    ----------
    items: :class:`str`
        A string listing elements.
    separator: :class:`str`
        The separator between each item. A comma by default.
    last_separator: :class:`str`
        The separator used for the last item. ' and ' by default.

    Returns
    -------
    :class:`list` of :class:`str`
        A list containing each one of the items.
    """
    if items is None:
        return None
    items = items.split(separator)
    last_item = items[-1]
    last_split = last_item.split(last_separator)
    if len(last_split) > 1:
        items[-1] = last_separator.join(last_split[:-1])
        items.append(last_split[-1])
    return [e.strip() for e in items]


def _recursive_strip(value):  # pragma: no cover
    if isinstance(value, dict):
        return {k: _recursive_strip(v) for k, v in value.items()}
    if isinstance(value, list):
        return [_recursive_strip(i) for i in value]
    if isinstance(value, str):
        return value.strip()
    return value


def deprecated(instead=None):  # pragma: no cover
    def actual_decorator(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            warnings.simplefilter('always', DeprecationWarning)
            if instead:
                fmt = "{0.__name__} is deprecated, use {1} instead."
            else:
                fmt = '{0.__name__} is deprecated.'

            warnings.warn(fmt.format(func, instead), stacklevel=3, category=DeprecationWarning)
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)
        return decorated
    return actual_decorator


def parse_popup(popup_content) -> Tuple[str, bs4.BeautifulSoup]:
    """Parse the information popups used through Tibia.com.

    Parameters
    ----------
    popup_content: :class:`str`
        The raw content of the javascript function that creates the popup.

    Returns
    -------
    :class:`str`
        The popup's title.
    :class:`bs4.BeautifulSoup`
        The parsed HTML content of the popup.
    """
    parts = popup_content.split(",", 2)
    title = parts[1].replace(r"'", "").strip()
    html = parts[-1].replace(r"\'", '"').replace(r"'", "").replace(",);", "").replace(", );", "").strip()
    parsed_html = bs4.BeautifulSoup(html, 'lxml')
    return title, parsed_html


results_pattern = re.compile(r'Results: ([\d,]+)')
page_pattern = re.compile(r'page=(\d+)')


def parse_pagination(pagination_block) -> Tuple[int, int, int]:
    """Parse a pagination section in Tibia.com and extracts its information.

    Parameters
    ----------
    pagination_block: :class:`bs4.Tag`
        The HTML containing the pagination information.

    Returns
    -------
    page : :class:`int`
        The current page.
    total_pages : :class:`int`
        The total number of pages.
    results_count : :class:`int`
        The total number of results.
    """
    pages_div, results_div = pagination_block.find_all("div")
    current_page_link = pages_div.find("span", {"class": "CurrentPageLink"})
    page_links = pages_div.find_all("span", {"class": "PageLink"})
    # pages_with_links = pages_div.select(#)
    first_or_last_pages = pages_div.find_all("span", {"class": "FirstOrLastElement"})
    page = -1
    total_pages = -1
    if first_or_last_pages:
        last_page_link = first_or_last_pages[-1].find("a")
        if last_page_link:
            m = page_pattern.search(last_page_link["href"])
            if m:
                total_pages = int(m.group(1))
        else:
            last_page_link = page_links[-2].find("a")
            total_pages = int(last_page_link.text) + 1
    else:
        last_page_link = page_links[-1]
        total_pages = int(last_page_link.text)
    try:
        page = int(current_page_link.text)
    except ValueError:
        if "First" in current_page_link.text:
            page = 1
        else:
            page = total_pages
    results_count = parse_integer(results_pattern.search(results_div.text).group(1))
    return page, total_pages, results_count



import datetime
import json
import unittest

from tests.tests_tibiapy import TestCommons
from tibiapy import House, HousesSection, InvalidContent, HouseEntry
from tibiapy.enums import HouseOrder, HouseStatus, HouseType

FILE_HOUSE_FULL = "house/tibiacom_full.txt"
FILE_HOUSE_STATUS_TRANSFER = "house/tibiacom_status_transfer.txt"
FILE_HOUSE_STATUS_NO_BIDS = "house/tibiacom_status_no_bids.txt"
FILE_HOUSE_STATUS_WITH_BIDS = "house/tibiacom_status_with_bids.txt"
FILE_HOUSE_STATUS_RENTED = "house/tibiacom_status_rented.txt"
FILE_HOUSE_NOT_FOUND = "house/tibiacom_not_found.txt"
FILE_HOUSE_LIST = "house/tibiacom_list.txt"
FILE_HOUSE_LIST_NOT_FOUND = "house/tibiacom_list_not_found.txt"
FILE_HOUSE_LIST_EMPTY = "house/tibiacom_list_empty.txt"


class TestsHouse(TestCommons, unittest.TestCase):
    def test_house_from_content(self):
        """Testing parsing a house"""
        content = self.load_resource(FILE_HOUSE_FULL)
        house = House.from_content(content)

        self.assertIsInstance(house, House)
        self.assertEqual(house.name, "Sorcerer's Avenue Labs 2e")
        self.assertEqual(house.beds, 1)
        self.assertTrue(house.rent, 715)
        self.assertEqual(house.status, HouseStatus.AUCTIONED)
        self.assertEqual(house.url, House.get_url(house.id, house.world))
        self.assertIsNone(house.owner)
        self.assertIsNone(house.owner_url)
        self.assertIsNotNone(house.highest_bidder)
        self.assertIsNotNone(house.highest_bidder_url)
        self.assertEqual(house.highest_bid, 0)

        house_json_raw = house.to_json()
        house_json = json.loads(house_json_raw)
        self.assertEqual(house.image_url, house_json["image_url"])

    def test_house_from_content_transferred(self):
        """Testing parsing a house being transferred"""
        house = House("Name")
        content = self.load_resource(FILE_HOUSE_STATUS_TRANSFER)
        house._parse_status(content)
        self.assertEqual(house.status, HouseStatus.RENTED)
        self.assertEqual(house.owner, "Xenaris mag")
        self.assertEqual(house.transferee, "Ivarr Bezkosci")
        self.assertIsNotNone(house.transferee_url)
        self.assertTrue(house.transfer_accepted)
        self.assertEqual(house.transfer_price, 850000)

    def test_house_parse_status_rented(self):
        """Testing parsing a rented status"""
        house = House("Name")
        content = self.load_resource(FILE_HOUSE_STATUS_RENTED)
        house._parse_status(content)
        self.assertEqual(house.status, HouseStatus.RENTED)
        self.assertEqual(house.owner, "Thorcen")
        self.assertIsInstance(house.paid_until, datetime.datetime)

    def test_house_parse_status_with_bids(self):
        """Testing parsing a house status with bids"""
        house = House("Name")
        content = self.load_resource(FILE_HOUSE_STATUS_WITH_BIDS)
        house._parse_status(content)
        self.assertEqual(house.status, HouseStatus.AUCTIONED)
        self.assertIsNone(house.owner)
        self.assertEqual(house.highest_bid, 15000)
        self.assertEqual(house.highest_bidder, "King of Bosnia")
        self.assertIsInstance(house.auction_end, datetime.datetime)

    def test_house_parse_status_without_bids(self):
        """Testing parsing the status of a house with no bids"""
        house = House("Name")
        content = self.load_resource(FILE_HOUSE_STATUS_NO_BIDS)
        house._parse_status(content)
        self.assertEqual(house.status, HouseStatus.AUCTIONED)
        self.assertIsNone(house.auction_end)

    def test_house_from_content_not_found(self):
        """Testing parsing a house that doesn't exist"""
        content = self.load_resource(FILE_HOUSE_NOT_FOUND)
        house = House.from_content(content)

        self.assertIsNone(house)

    def test_house_from_content_unrelated(self):
        """Testing parsing an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            House.from_content(content)

    def test_house_section_from_content(self):
        """Testing parsing the house list of a world and city"""
        content = self.load_resource(FILE_HOUSE_LIST)
        house_results = HousesSection.from_content(content)

        self.assertIsInstance(house_results, HousesSection)
        self.assertEqual("Carlin", house_results.town)
        self.assertEqual("Alumbra", house_results.world)
        self.assertEqual(HouseStatus.RENTED, house_results.status)
        self.assertEqual(HouseType.HOUSE, house_results.house_type)
        self.assertEqual(HouseOrder.RENT, house_results.order)

        houses = house_results.entries
        self.assertIsInstance(houses, list)
        self.assertGreater(len(houses), 0)
        self.assertIsInstance(houses[0], HouseEntry)
        self.assertEqual(houses[0].town, "Carlin")
        self.assertEqual(houses[0].type, HouseType.HOUSE)
        self.assertEqual(houses[0].status, HouseStatus.RENTED)
        self.assertIsInstance(houses[0].id, int)

        self.assertEqual(houses[25].status, HouseStatus.RENTED)

    def test_house_section_from_content_empty(self):
        """Testing parsing an empty house list"""
        content = self.load_resource(FILE_HOUSE_LIST_EMPTY)
        house_results = HousesSection.from_content(content)

        self.assertIsInstance(house_results, HousesSection)
        self.assertEqual("Edron", house_results.town)
        self.assertEqual("Antica", house_results.world)
        self.assertEqual(HouseStatus.AUCTIONED, house_results.status)
        self.assertEqual(HouseType.GUILDHALL, house_results.house_type)
        self.assertEqual(HouseOrder.RENT, house_results.order)
        self.assertEqual(len(house_results.entries), 0)

    def test_house_section_from_content_not_found(self):
        """Testing parsing an empty house list"""
        content = self.load_resource(FILE_HOUSE_LIST_NOT_FOUND)
        results = HousesSection.from_content(content)
        self.assertIsInstance(results, HousesSection)
        self.assertEqual(0, len(results.entries))

    def test_house_section_from_content_unrelated(self):
        """Testing parsing an unrelated section"""
        content = self.load_resource(self.FILE_UNRELATED_SECTION)
        with self.assertRaises(InvalidContent):
            HousesSection.from_content(content)

